"""
This module is designed to make STCAL tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


STCAL copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.


Original: https://github.com/spacetelescope/stcal/blob/main/src/stcal/tweakreg/astrometric_utils.py
"""

import os

import requests
from astropy import table
from astropy import units as u
from astropy.coordinates import SkyCoord
from astropy.table import Table

from spike.stcal.alignment_util import compute_fiducial

ASTROMETRIC_CAT_ENVVAR = "ASTROMETRIC_CATALOG_URL"
DEF_CAT_URL = "http://gsss.stsci.edu/webservices"

SERVICELOCATION = os.environ.get(ASTROMETRIC_CAT_ENVVAR, DEF_CAT_URL)

TIMEOUT = 30.0  # in seconds

"""

Primary function for creating an astrometric reference catalog.

"""


__all__ = [
    "TIMEOUT",
    "compute_radius",
    "create_astrometric_catalog",
    "get_catalog"]


def create_astrometric_catalog(
        wcs,
        epoch,
        catalog="GAIADR3",
        output="ref_cat.ecsv",
        gaia_only=False,
        table_format="ascii.ecsv",
        num_sources=None):
    """Create an astrometric catalog that covers the inputs' field-of-view.

    Parameters
    ----------
    wcs : `~astropy.wcs.WCS`
        WCS object specified by the user as generated by
        `resample.resample_utils.make_output_wcs`. This will typically
        have the same plate-scale and orientation as the first member in the
        list of input images to make_output_wcs. Fortunately, for alignment,
        this doesn't matter since no resampling of data will be performed.

    epoch : float
        Reference epoch used to update the coordinates for proper motion
        (in decimal year).

    catalog : str, optional
        Name of catalog to extract astrometric positions for sources in the
        input images' field-of-view. Default: GAIADR3. Options available are
        documented on the catalog web page.

    output : str, optional
        Filename to give to the astrometric catalog read in from the master
        catalog web service.  If None, no file will be written out.

    gaia_only : bool, optional
        Specify whether or not to only use sources from GAIA in output catalog

    num_sources : int
        Maximum number of brightest/faintest sources to return in catalog.
        If `num_sources` is negative, return that number of the faintest
        sources.  By default, all sources are returned.



    Notes
    -----
    This function will point to astrometric catalog web service defined
    through the use of the ASTROMETRIC_CATALOG_URL environment variable.

    Returns
    -------
    ref_table : `~astropy.table.Table`
        Astropy Table object of the catalog
    """
    # start by creating a composite field-of-view for all inputs
    radius, fiducial = compute_radius(wcs)

    # perform query for this field-of-view
    ref_dict = get_catalog(
        fiducial[0],
        fiducial[1],
        epoch=epoch,
        search_radius=radius,
        catalog=catalog)
    if len(ref_dict) == 0:
        return ref_dict

    colnames = ("ra", "dec", "mag", "objID", "epoch")
    ref_table = ref_dict[colnames]

    # Add catalog name as meta data
    ref_table.meta["catalog"] = catalog
    ref_table.meta["gaia_only"] = gaia_only

    # rename coordinate columns to be consistent with tweakwcs
    ref_table.rename_column("ra", "RA")
    ref_table.rename_column("dec", "DEC")

    # Append GAIA ID as a new column to the table...
    gaia_sources = []
    for source in ref_dict:
        if "GAIAsourceID" in source:
            g = source["GAIAsourceID"]
            if gaia_only and g.strip() == "":
                continue
        else:
            g = "-1"  # indicator for no source ID extracted
        gaia_sources.append(g)

    gaia_col = table.Column(data=gaia_sources, name="GaiaID", dtype="U25")
    ref_table.add_column(gaia_col)

    # sort table by magnitude, fainter to brightest
    ref_table.sort("mag", reverse=True)

    # If specified by the use through the 'num_sources' parameter,
    # trim the returned catalog down to the brightest 'num_sources' sources
    # Should 'num_sources' be a negative value, it will return the faintest
    # 'num_sources' sources.
    if num_sources is not None:
        indx = -1 * num_sources
        ref_table = ref_table[:indx] if num_sources < 0 else ref_table[indx:]

    # Write out table to a file, if specified
    if output is not None:
        ref_table.write(output, format=table_format, overwrite=True)

    return ref_table


def compute_radius(wcs):
    """Compute the radius from the center to the furthest edge of the WCS."""
    fiducial = compute_fiducial([wcs,])
    img_center = SkyCoord(
        ra=fiducial[0] * u.degree,
        dec=fiducial[1] * u.degree)
    wcs_foot = wcs.footprint()
    img_corners = SkyCoord(ra=wcs_foot[:, 0] * u.degree,
                           dec=wcs_foot[:, 1] * u.degree)
    radius = img_center.separation(img_corners).max().value

    return radius, fiducial


def get_catalog(
    right_ascension,
    declination,
    epoch=2016.0,
    search_radius=0.1,
    catalog="GAIADR3",
    timeout=TIMEOUT,
):
    """Extract catalog from VO web service.

    Parameters
    ----------
    right_ascension : float
        Right Ascension (RA) of center of field-of-view (in decimal degrees)

    declination : float
        Declination (Dec) of center of field-of-view (in decimal degrees)

    epoch : float, optional
        Reference epoch used to update the coordinates for proper motion
        (in decimal year). Default: 2016.0

    search_radius : float, optional
        Search radius (in decimal degrees) from field-of-view center to use
        for sources from catalog.  Default: 0.1 degrees

    catalog : str, optional
        Name of catalog to query, as defined by web-service. Default: 'GAIADR3'

    timeout : float, optional
        Timeout in seconds to wait for the catalog web service to respond. Default: 30.0 s

    Returns
    -------
    csv : `~astropy.table.Table`
        CSV object of returned sources with all columns as provided by catalog

    """
    service_type = "vo/CatalogSearch.aspx"
    spec_str = "RA={}&DEC={}&EPOCH={}&SR={}&FORMAT={}&CAT={}&MINDET=5"
    headers = {"Content-Type": "text/csv"}
    fmt = "CSV"

    spec = spec_str.format(
        right_ascension,
        declination,
        epoch,
        search_radius,
        fmt,
        catalog
    )
    service_url = f"{SERVICELOCATION}/{service_type}?{spec}"
    try:
        rawcat = requests.get(service_url, headers=headers, timeout=timeout)
    except requests.exceptions.ConnectionError:
        raise requests.exceptions.ConnectionError(
            "Could not connect to the VO API server. Try again later."
        )
    except requests.exceptions.Timeout:
        raise requests.exceptions.Timeout("The request to the VO API server timed out.")
    except requests.exceptions.RequestException:
        raise requests.exceptions.RequestException(
            "There was an unexpected error with the request."
        )
    r_contents = rawcat.content.decode()  # convert from bytes to a String
    rstr = r_contents.split("\r\n")
    # remove initial line describing the number of sources returned
    # CRITICAL to proper interpretation of CSV data
    del rstr[0]

    return Table.read(rstr, format="csv")


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/lib/diff.py
"""

from collections import Counter, UserList, defaultdict
from copy import copy
import logging
from pathlib import Path
import re

from spike.jwstcal.associations_lib_product_utils import sort_by_candidate
from spike.jwstcal.associations_load_asn import load_asn
from spike.jwstcal.lib_suffix import remove_suffix

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

__all__ = [
    'compare_asn_files',
    'compare_asn_lists',
    'compare_asns',
    'compare_membership',
    'compare_product_membership',
]

# #########################
# Define the types of diffs
# #########################
class DiffError(AssertionError):
    """Base Class for difference errors

    Attributes
    ----------
    args : tuple
        Standard exception arguments

    asns : [Association[,...]]
        List of associations that generated the exception
    """
    def __init__(self, *args, asns=None, **kwargs):
        super().__init__(*args, **kwargs)
        if asns is None:
            asns = list()
        self.asns = asns


class CandidateLevelError(DiffError):
    """Candidate level mismatch"""


class DuplicateMembersError(DiffError):
    """Duplicate members within a product"""


class DuplicateProductError(DiffError):
    """Duplicate products found"""


class DifferentProductSetsError(DiffError):
    """Different product sets between groups of associations"""


class MemberLengthDifference(DiffError):
    """Difference in number of members between associations"""


class MemberMismatchError(DiffError):
    """Membership does not match"""


class SubsetError(DiffError):
    """One product is a subset of another"""


class TypeMismatchError(DiffError):
    """Association type mismatch"""


class UnaccountedMembersError(DiffError):
    """Members not present in other"""


class MultiDiffError(UserList, DiffError):
    """List of diff errors"""
    def __init__(self, *args, **kwargs):
        super(MultiDiffError, self).__init__(*args, **kwargs)

    @property
    def asns(self):
        """Return the list of affected associations for all contained errors"""
        asns = list()
        for err in self:
            asns.extend(err.asns)
        return asns

    @property
    def err_types(self):
        """Return list of error types"""
        err_types = [type(err) for err in self]
        return err_types

    def __str__(self):
        message = ['Following diffs found:\n']
        for diff in self:
            message.extend([
                '\n****\n', str(diff), '\n'
            ])
        return ''.join(message)


def compare_asn_files(left_paths, right_paths):
    """Compare association files

    Parameters
    ----------
    left_paths: [Path or str[, ...]]
        Set of association files

    right_paths: [Path or str[, ...]]
        Set of association files to compare against

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    __tracebackhide__ = True
    # Read in all the associations, separating out the products into separate associations.
    left_asns = []
    for path in left_paths:
        with open(path, 'r') as fh:
            asn = load_asn(fh)
        single_product_asns = separate_products(asn)

        # Collect the associations.
        left_asns += single_product_asns

    right_asns = []
    for path in right_paths:
        with open(path, 'r') as fh:
            asn = load_asn(fh)
        single_product_asns = separate_products(asn)

        # Collect the associations.
        right_asns += single_product_asns

    # Compare the associations.
    compare_asn_lists(left_asns, right_asns)


def compare_asn_lists(left_asns, right_asns):
    """Compare to lists of associations

    Both association lists must contain associations that only have
    single products. Use `separate_products` prior to calling this
    function.

    Parameters
    ----------
    left_asns: [`Association`[, ...]]
        Group of associations

    right_asns: [`Association`[, ...]]
        Group of associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    __tracebackhide__ = True
    diffs = MultiDiffError()

    # Ensure that product names are unique
    left_product_names, left_duplicates = get_product_names(left_asns)
    right_product_names, right_duplicates = get_product_names(right_asns)
    if left_duplicates:
        try:
            check_duplicate_products(left_asns, product_names=left_product_names, dup_names=left_duplicates)
        except MultiDiffError as dup_errors:
            diffs.extend(dup_errors)
    if right_duplicates:
        try:
            check_duplicate_products(right_asns, product_names=right_product_names, dup_names=right_duplicates)
        except MultiDiffError as dup_errors:
            diffs.extend(dup_errors)

    # Ensure that the product name lists are the same.
    name_diff = left_product_names ^ right_product_names
    if name_diff:
        diffs.append(DifferentProductSetsError(
            'Left and right associations do not share a common set of products: {}'
            ''.format(name_diff)
        ))

    # Compare like product associations
    left_asns_by_product = {
        asn['products'][0]['name']: asn
        for asn in left_asns
    }
    right_asns_by_product = {
        asn['products'][0]['name']: asn
        for asn in right_asns
    }
    for product_name in left_product_names:
        try:
            compare_asns(left_asns_by_product[product_name], right_asns_by_product[product_name])
        except MultiDiffError as compare_diffs:
            diffs.extend(compare_diffs)
        except KeyError:
            # Most likely due to a previous error. Ignore
            pass

    if diffs:
        raise diffs


def compare_asns(left, right):
    """Compare two associations

    This comparison will include metadata such as
    `asn_type` and membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there is a difference.
    """
    _compare_asns(left, right)


def _compare_asns(left, right):
    """Compare two associations

    This comparison will include metadata such as
    `asn_type` and membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.

    Note
    ----
    This comparison is dependent on the associations being JWST-like associations.
    The attributes that are compared are as follows:

        - key `asn_type`
        - key `products`. Specifically the following are compared:
            - Length of the list
            - key `name` for each product
            - key `members` for each product
        - For the member lists of each product, the following are compared:
            - Length of the list
            - key `expname` for each member
            - key 'exptype` for each member
    """
    diffs = MultiDiffError()

    # Assert that the same result type is the same.
    if left['asn_type'] != right['asn_type']:
        diffs.append(TypeMismatchError(
            'Type mismatch {} != {}'.format(left['asn_type'], right['asn_type'])
        ))

    # Assert that the level of association candidate is the same.
    # Cannot guarantee value, but that the 'a'/'c'/'o' levels are similar.
    if left['asn_id'][0] != right['asn_id'][0]:
        diffs.append(CandidateLevelError(
            f"Candidate level mismatch left '{left['asn_id'][0]}' != right '{right['asn_id'][0]}'"
        ))

    # Membership
    try:
        compare_membership(left, right)
    except MultiDiffError as compare_diffs:
        diffs.extend(compare_diffs)

    if diffs:
        raise diffs


def compare_membership(left, right):
    """Compare two associations' membership

    Parameters
    ---------
    left, right : dict
        Two, individual, associations to compare

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    diffs = MultiDiffError()
    products_left = left['products']
    products_right = copy(right['products'])

    if len(products_left) != len(products_right):
        diffs.append(DifferentProductSetsError(
            '# products differ: {left_len} != {right_len}'
            ''.format(left_len=len(products_left), right_len=len(products_right))
        ))

    for left_idx, left_product in enumerate(products_left):
        left_product_name = components(left_product['name'])
        for right_idx, right_product in enumerate(products_right):
            if components(right_product['name']) != left_product_name:
                continue
            try:
                compare_product_membership(left_product, right_product, strict_expname=False)
            except MultiDiffError as compare_diffs:
                diffs.extend(compare_diffs)
            products_right.remove(right_product)
            break
        else:
            diffs.append(DifferentProductSetsError(
                'Left has {n_products} left over'
                ''.format(n_products=len(products_left))
            ))


    if len(products_right) != 0:
        diffs.append(DifferentProductSetsError(
            'Right has {n_products} left over'
            ''.format(n_products=len(products_right))
        ))

    if diffs:
        raise diffs


def compare_product_membership(left, right, strict_expname=True):
    """Compare membership between products

    If the membership is exactly the same, or other special conditions,
    no error is raised.

    Otherwise, the following errors will be raised:

    - DuplicateMemberError
      A member exists multiple times in the member list.

    - MemberLengthDifference
      Number of members in the two products differ.

    - MemberMismatchError
      Members with the same `expname` do not share the other attributes.

    - SubsetError
      A member list is a subset of another member list.

    - UnaccountedMembersError
      Members exist in one association that do not exist in the other.

    Parameters
    ---------
    left, right : dict
        Two, individual, association products to compare. Note that
        these are not associations, just products from associations.

    strict_expname : bool
        Compare `expname` exactly. If False, `expname` munging
        will occur. See `exposure_name` for further details.
        Generally False for when comparing unreleated association lists.
        Generally True when comparing related associations;
        those associations generated together.

    Raises
    ------
    MultiDiffError
        If there are differences. The message will contain
        all the differences.
    """
    diffs = MultiDiffError()

    # Determine how to compare expnames
    if strict_expname:
        def munge_expname(expname): 
            return expname
    else:
        munge_expname = exposure_name

    # Check for duplicate members.
    try:
        check_duplicate_members(left)
    except DuplicateMembersError as dup_member_error:
        diffs.append(dup_member_error)
    try:
        check_duplicate_members(right)
    except DuplicateMembersError as dup_member_error:
        diffs.append(dup_member_error)

    if len(right['members']) != len(left['members']):
        diffs.append(MemberLengthDifference(
            'Product Member length differs:'
            ' Left Product {left_product_name} len {left_len} !=  '
            ' Right Product {right_product_name} len {right_len}'
            ''.format(left_product_name=left['name'], left_len=len(left['members']),
                      right_product_name=right['name'], right_len=len(right['members']))
        ))

    members_right = copy(right['members'])
    left_unaccounted_members = []
    for left_member in left['members']:
        for right_member in members_right:
            if munge_expname(left_member['expname']) != munge_expname(right_member['expname']):
                continue

            if left_member['exptype'] != right_member['exptype']:
                diffs.append(MemberMismatchError(
                    'Left {left_expname}:{left_exptype}'
                    ' != Right {right_expname}:{right_exptype}'
                    ''.format(left_expname=left_member['expname'], left_exptype=left_member['exptype'],
                              right_expname=right_member['expname'], right_exptype=right_member['exptype'])
                ))

            members_right.remove(right_member)
            break
        else:
            left_unaccounted_members.append(left_member)

    if len(left_unaccounted_members):
        diffs.append(UnaccountedMembersError(
            f'Left has {len(left_unaccounted_members)} unaccounted members. Members are {left_unaccounted_members}'
        ))

    if len(members_right) != 0:
        diffs.append(UnaccountedMembersError(
            f'Right has {len(members_right)} unaccounted members. Members are {members_right}'
        ))

    # Check if one is a subset of the other.
    err_types = diffs.err_types
    is_subset = (len(diffs) == 2) and (MemberLengthDifference in err_types) and (UnaccountedMembersError in err_types)
    if is_subset:
        diffs = MultiDiffError([SubsetError(f'Products are subsets: {left["name"]} {right["name"]}')])

    if diffs:
        raise diffs


def check_duplicate_members(product):
    """Check for duplicate members in an association product

    The check is based solely on `expname`.

    Parameters
    ----------
    product : dict
        Association product to check.

    Raises
    ------
    MultiDiffError
        If the product has duplicate members.
    """
    seen = set()
    dups = []
    for expname in [member['expname'] for member in product['members']]:
        if expname in seen:
            dups.append(expname)
        else:
            seen.add(expname)

    if dups:
        raise DuplicateMembersError(
            f'Product {product["name"]} has duplicate members {dups}'
        )


def check_duplicate_products(asns, product_names=None, dup_names=None):
    """Check for duplicate products in a list of associations

    Duplicate products are defined as any products that share the same name.
    The errors flagged are listed below. If no `MultiDiffError` is raised,
    There are no duplicate products.

    - DuplicateProductError
      The general error for two products that share name and otherwise do
      not fall into any other category

    - SubsetError
      When one or more products are full subsets of another.

    There is a special case where products may have the same name. For Level 2
    processing, exposures can be processed either as time-series, indicated by
    the "rateints" suffix, or as single exposures, indicated by the "rate" suffix.
    Usually any particular exposure is treated one way or the other. However, for
    coronagraphic data, exposures are processed in both ways. In this case, the product
    name and membership will be the same, except that the members will have the different
    suffixes.

    Parameters
    ----------
    asns : [Associations[,...]]
        The associations to compare. Each association should only have one
        product. Use `separate_products` prior to calling if necessary.

    product_names : [str[,...]]
        Product names in given associations.
        If None, will be generated internally.

    dup_names : [str[,...]]
        Duplicate product names in the given associations.
        If None, will be generated internally.

    Raises
    ------
    MultiDiffError
    """
    if product_names is None or dup_names is None:
        product_names, dup_names = get_product_names(asns)
    if not dup_names:
        return

    # Order associations by candidate. This will allow choosing which
    # association should be considered "primary" when duplicates are found.
    # Categorize by duplicated product names.
    ordered_asns = sort_by_candidate(asns)
    asn_by_product = defaultdict(list)
    for asn in ordered_asns:
        asn_by_product[asn['products'][0]['name']].append(asn)

    diffs = MultiDiffError()
    for product in dup_names:
        dup_asns = asn_by_product[product]
        current_asn = dup_asns.pop()
        for asn in dup_asns:
            try:
                compare_product_membership(current_asn['products'][0], asn['products'][0])
            except MultiDiffError as compare_diffs:
                # Re-check membership, but allow products that different only in the suffix
                # of the members. If there are no differences, then the products are not duplicates.
                try:
                    compare_product_membership(current_asn['products'][0], asn['products'][0], strict_expname=False)
                except MultiDiffError:
                    # Not interested if the second compare generates diffs.
                    # The initial diffs will be more informative.
                    diffs.extend(compare_diffs)
            else:
                # Associations are exactly the same. Pure duplicate.
                diffs.append(DuplicateProductError(
                    f'Associations share product name {product}',
                    asns=[current_asn, asn]
                ))
    if diffs:
        raise diffs


# #########
# Utilities
# #########
def components(s):
    """split string into its components"""
    return set(re.split('[_-]', s))


def exposure_name(path):
    """Extract the exposure name from a Stage 2 file name

    Parameters
    ----------
    path : Path or str
        The file name or path.

    Returns
    -------
    exposure : str
        The exposure name
    """
    path = Path(path)
    exposure, _ = remove_suffix(path.stem)
    return exposure


def get_product_names(asns):
    """Return product names from associations and flag duplicates

    Parameters
    ----------
    asns: [`Association`[, ...]]

    Returns
    -------
    product_names, duplicates: set(str[, ...]), [str[,...]]
        2-tuple consisting of the set of product names and the list of duplicates.
    """
    product_names = [
        asn['products'][0]['name']
        for asn in asns
    ]

    dups = [
        name
        for name, count in Counter(product_names).items()
        if count > 1
    ]
    if dups:
        logger.debug(
            'Duplicate product names: %s', dups
        )

    return set(product_names), dups


def separate_products(asn):
    """Separate products into individual associations

    Parameters
    ----------
    asn: `Association`
        The association to split

    Returns
    -------
    separated: [`Association`[, ...]]
        The list of separated associations
    """
    separated = []
    for product in asn['products']:
        new_asn = copy(asn)
        new_asn['products'] = [product]
        separated.append(new_asn)
    return separated


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/model_blender/_tablebuilder.py
"""

import numpy as np


class _MissingValueType:
    pass

_MISSING_VALUE = _MissingValueType()


def _convert_dtype(value):
    """Convert numarray column dtype into YAML-compatible format description"""
    if 'U' in value:
        # working with a string description
        str_len = int(value[value.find('U') + 1:])
        new_dtype = ['ascii', str_len]
    elif value == 'bool':
        # cast all bool to int8 to avoid issues on write
        new_dtype = 'int8'
    else:
        new_dtype = str(value)

    return new_dtype


def table_to_schema(table):
    """
    Construct a schema for a table.

    Convert a "table" (a structured ndarray) to a stdatamodels
    sub-schema that will allow the "table" to be stored to a fits
    extension HDRTAB.

    Parameters
    ----------
    table: ndarray
        The structured array containing the data (and datatype).

    Returns
    -------
    subschema: dict
        stdatamodels for the "table" datatype.
    """
    return {
        'title': 'Combined header table',
        'fits_hdu': 'HDRTAB',
        'datatype': [
            {
                'name': col_name,
                'datatype': _convert_dtype(str(table.dtype[col_name])),
            }
            for col_name in table.dtype.fields
        ],
    }


class TableBuilder:
    """
    Class to build a metadata table.

    Used to incrementally build a metadata "table" (a numpy
    structured array) containing metadata from several models.

    >>> tb = TableBuilder({"meta.filename": "FN"})
    >>> tb.header_to_row({"meta.filename": "foo.fits"})
    >>> tb.build_table()
    rec.array([('foo.fits',)],
         dtype=[('FN', '<U8')])
    """
    def __init__(self, attr_to_column):
        """
        Create a new `TableBuilder`.

        Parameters
        ----------
        attr_to_column: dict
            A one-to-one mapping of attribute names (as
            dotted paths like "meta.filename") to column
            names (strings).
        """
        self.attr_to_column = attr_to_column
        self.columns = {col: [] for col in self.attr_to_column.values()}
        if len(attr_to_column) != len(self.columns):
            raise ValueError(f"Invalid attr_to_column, mapping is not 1-to-1: {attr_to_column}")

    def header_to_row(self, header):
        """
        Add metadata in header to the table.

        This function will add a complete row for each
        header. If header is missing a required attribute
        (as defined in the mapping provided to
        `TableBuilder.__init__`) the column with this
        missing value for this row will contain a ``nan``.

        Parameters
        ----------
        header: dict
            Often produced from ``Datamodel.to_flat_dict``.
        """
        row = {}
        for attr in self.attr_to_column:
            if attr in header:
                row[attr] = header[attr]
        self._add_row(row)

    def _add_row(self, row):
        for attr, col in self.attr_to_column.items():
            self.columns[col].append(row[attr] if attr in row else _MISSING_VALUE)

    def build_table(self):
        """
        Build the final "table".

        Returns
        -------
        table: numpy.ndarray
            Structured array containing fields with datatypes
        """
        arrays = []
        table_dtype = []
        for col, items in self.columns.items():
            if all((i is _MISSING_VALUE for i in items)):
                continue
            arrays.append(np.array([np.nan if i is _MISSING_VALUE else i for i in items]))
            table_dtype.append((col, arrays[-1].dtype))
        return np.rec.fromarrays(arrays, dtype=table_dtype)


from spike import psf


############################################## HST #############################################
acs_path = '/path/to/acs/data'
## files contained in directory:
# j8pu42ecq_flc.fits
# j8pu42egq_flc.fits
# j8pu42esq_flc.fits
# j8pu42evq_flc.fits

### test TinyTim output ###
psf.hst(img_dir = acs_path, obj = '10:00:33.0178 +02:09:52.304', img_type = 'flc', 
	inst = 'ACS', camera = 'WFC', method='TinyTim', savedir = 'psfs_tiny', verbose = True,
	pretweaked = False)

# ### test TinyTim (Gillis et al. mod) output ###
psf.hst(img_dir = acs_path, obj = '10:00:33.0178 +02:09:52.304', img_type = 'flc', 
	inst = 'ACS', camera = 'WFC', method='TinyTim_Gillis', savedir = 'psfs_tinygillis', verbose = True,
	pretweaked = True)

# ### test STDPSF output ###
psf.hst(img_dir = acs_path, obj = '10:00:33.0178 +02:09:52.304', img_type = 'flc', 
	inst = 'ACS', camera = 'WFC', method='stdpsf', savedir = 'psfs_stdpsf', verbose = True,
	pretweaked = True)

# ### test ePSF output ###
psf.hst(img_dir = acs_path, obj = '10:00:33.0178 +02:09:52.304', img_type = 'flc', 
	inst = 'ACS', camera = 'WFC', method='epsf', savedir = 'psfs_epsf', verbose = True,
	pretweaked = True, thresh = 200)

# ### test PSFEx output ###
psf.hst(img_dir = acs_path, obj = '10:00:33.0178 +02:09:52.304', img_type = 'flc', 
	inst = 'ACS', camera = 'WFC', method='PSFEx', savedir = 'psfs_psfex', verbose = True,
	pretweaked = True)



############################################# JWST #############################################
nircam_path = '/path/to/nircam/data'
## files contained in directory:
# jw02514162001_03201_00001_nrca2_cal.fits
# jw02514162001_03201_00002_nrca2_cal.fits
# jw02514162001_03201_00003_nrca2_cal.fits

## test WebbPSF output ###
psf.jwst(img_dir = nircam_path, obj = '10:00:31.432 +02:10:26.29', img_type = 'cal',
	inst = 'NIRCam', method='webbpsf', savedir = 'psfs_webbpsf', verbose = True,
	pretweaked = False)

### test STDPSF output ###
psf.jwst(img_dir = nircam_path, obj = '10:00:31.432 +02:10:26.29', img_type = 'tweakregstep', 
	inst = 'NIRCam', method='stdpsf', savedir = 'psfs_jwst_stdpsf', verbose = True,
	pretweaked = True)

### test ePSF output ###
psf.jwst(img_dir = nircam_path, obj = '10:00:31.432 +02:10:26.29', img_type = 'tweakregstep', 
	inst = 'NIRCam', method='epsf', savedir = 'psfs_jwst_epsf', verbose = True,
	pretweaked = True, starselect = 'IRAF', thresh = 3)

### test PSFEx output ###
psf.jwst(img_dir = nircam_path, obj = '10:00:31.432 +02:10:26.29', img_type = 'tweakregstep', 
	inst = 'NIRCam', method='PSFEx', savedir = 'psfs_jwst_psfex', verbose = True,
	pretweaked = True)


################################# code to generate figures #####################################
import matplotlib.pyplot as plt
import numpy as np
import albumpl.cmap
from astropy.io import fits


fig, ax = plt.subplots(1, 5, figsize = (20, 4), gridspec_kw = {'wspace':0.1})

tt = fits.open('psfs_tiny/150d08m15.267s+2d09m52.304s_F475W_psf_drz.fits')
ttim = tt[1].data[325:465, 940:1080]
ax[0].imshow(ttim, vmin = np.percentile(ttim, 20), vmax = np.percentile(ttim, 97), origin = 'lower', cmap = 'LondonCalling_r')
ax[0].text(10, 125, s = 'TinyTim', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[0].axis('off')

ttg = fits.open('psfs_tinygillis/150d08m15.267s+2d09m52.304s_F475W_psf_drz.fits')
ttgim = ttg[1].data[325:465, 940:1080]
ax[1].imshow(ttgim, vmin = np.percentile(ttgim, 20), vmax = np.percentile(ttgim, 97), origin = 'lower', cmap = 'LondonCalling_r')
ax[1].text(10, 125, s = 'TinyTim (Gillis+20)', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[1].axis('off')

std = fits.open('psfs_stdpsf/150d08m15.267s+2d09m52.304s_F475W_psf_drz.fits')
stdim = std[1].data[325:465, 940:1080]
# ax[2].imshow(stdim, vmin = np.percentile(stdim[stdim != 0], 5), vmax = np.percentile(stdim[stdim != 0], 60), origin = 'lower', cmap = 'LondonCalling_r')
ax[2].imshow(stdim, vmin = np.percentile(stdim, 20), vmax = np.percentile(stdim, 99), origin = 'lower', cmap = 'LondonCalling_r')
ax[2].text(10, 125, s = 'STDPSF', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[2].axis('off')

epsf = fits.open('psfs_epsf/150d08m15.267s+2d09m52.304s_F475W_psf_drz.fits')
epsfim = epsf[1].data[325:465, 940:1080]
ax[3].imshow(epsfim, vmin = np.percentile(epsfim, 20), vmax = np.percentile(epsfim, 97), origin = 'lower', cmap = 'LondonCalling_r')
ax[3].text(10, 125, s = 'ePSF', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[3].axis('off')


psfex = fits.open('psfs_psfex/150d08m15.267s+2d09m52.304s_F475W_psf_drz.fits')
psfexim_ = psfex[1].data[325:465, 940:1080]
ax[4].imshow(psfexim_, vmin = np.percentile(psfexim_[psfexim_ >0], 20), vmax = np.percentile(psfexim_[psfexim_ >0], 95), origin = 'lower', cmap = 'LondonCalling_r')
ax[4].text(10, 125, s = 'PSFEx', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[4].axis('off')

fig.savefig('spike_psf_hstcompare.png', bbox_inches = 'tight')


fig, ax = plt.subplots(1, 4, figsize = (16, 4), gridspec_kw = {'wspace':0.1})

jw = fits.open('psfs_webbpsf/100031+021026_F115W_resamplestep.fits')
jwim = jw[1].data[1114-70: 1114+70, 553-70:553+70]
ax[0].imshow(jwim, vmin = np.nanpercentile(jwim, 20), vmax = np.nanpercentile(jwim, 97), origin = 'lower', cmap = 'LondonCalling_r')
ax[0].text(10, 125, s = 'WebbPSF', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[0].axis('off')

std = fits.open('psfs_jwst_stdpsf/100031+021026_F115W_resamplestep.fits')
stdim = std[1].data[1114-70: 1114+70, 553-70:553+70]
ax[1].imshow(stdim, vmin = np.nanpercentile(stdim, 20), vmax = np.nanpercentile(stdim, 99), origin = 'lower', cmap = 'LondonCalling_r')
ax[1].text(10, 125, s = 'STDPSF', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[1].axis('off')

## this one has to be generated with different values than the defaults -- ePSF is very sensitive to threshold value
epsf = fits.open('psfs_jwst_epsf/100031+021026_F115W_resamplestep.fits')
epsfim = epsf[1].data[1114-70: 1114+70, 553-70:553+70]
ax[2].imshow(epsfim, vmin = np.nanpercentile(epsfim, 20), vmax = np.nanpercentile(epsfim, 97), origin = 'lower', cmap = 'LondonCalling_r')
ax[2].text(10, 125, s = 'ePSF', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[2].axis('off')


psfex = fits.open('psfs_jwst_psfex/100031+021026_F115W_resamplestep.fits')
psfexim_ = psfex[1].data[1114-70: 1114+70, 553-70:553+70]
ax[3].imshow(psfexim_, vmin = np.nanpercentile(psfexim_[psfexim_ >0], 20), vmax = np.nanpercentile(psfexim_[psfexim_ >0], 95), origin = 'lower', cmap = 'LondonCalling_r')
ax[3].text(10, 125, s = 'PSFEx', fontsize = 15, fontfamily = 'monospace', color = '#343530')
ax[3].axis('off')


fig.savefig('spike_psf_jwstcompare.png', bbox_inches = 'tight')





"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/lib/utilities.py
"""


import logging
from ast import literal_eval
from functools import wraps

from numpy.ma import masked

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def return_on_exception(exceptions=(Exception,), default=None):
    """Decorator to force functions raising exceptions to return a value

    Parameters
    ----------
    exceptions: (Exception(,...))
        Tuple of exceptions to catch

    default: obj
        The value to return when a specified exception occurs
    """

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except exceptions as err:
                logger.debug(
                    "Caught exception %s in function %s, forcing return value of %s",
                    err,
                    func,
                    default,
                )
                return default

        return wrapper

    return decorator


def evaluate(value):
    """Evaluate a value

    Parameters
    ----------
    value : str
        The string to evaluate.

    Returns
    -------
    type or str
        The evaluation. If the value cannot be
        evaluated, the value is simply returned
    """
    try:
        evaled = literal_eval(value)
    except (ValueError, SyntaxError):
        evaled = value
    return evaled


def getattr_from_list(adict, attributes, invalid_values=None):
    """Retrieve value from dict using a list of attributes

    Parameters
    ----------
    adict : dict
        dict to retrieve from

    attributes : list
        List of attributes

    invalid_values : set
        A set of values that essentially mean the
        attribute does not exist.

    Returns
    -------
    (attribute, value)
        Returns the value and the attribute from
        which the value was taken.

    Raises
    ------
    KeyError
        None of the attributes are found in the dict.
    """
    if invalid_values is None:
        invalid_values = set()

    for attribute in attributes:
        try:
            result = adict[attribute]
        except KeyError:
            continue
        else:
            if result is masked:
                continue
            if result not in invalid_values:
                return attribute, result
            else:
                continue
    else:
        raise KeyError(f"Object has no attributes in {attributes}")


@return_on_exception(exceptions=(KeyError,), default=None)
def getattr_from_list_nofail(*args, **kwargs):
    """Call getattr_from_list without allows exceptions.

    If the specified exceptions are caught, return `default`
    instead.

    Parameters
    ----------
    See `getattr_from_list`
    """
    return getattr_from_list(*args, **kwargs)


def is_iterable(obj):
    """General iterator check"""
    return (
        not isinstance(obj, str)
        and not isinstance(obj, tuple)
        and hasattr(obj, "__iter__")
    )


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/resample/resample.py
"""

import json
import logging
import os
from typing import List

import numpy as np
from astropy import units as u
from astropy.coordinates import SkyCoord
from drizzle import cdrizzle, util
from roman_datamodels import datamodels, maker_utils, stnode
from spike.stcal.alignment_util import compute_s_region_keyword, compute_scale

from spike.romancal.associations_asn_from_list import asn_from_list

import spike.romancal.assign_wcs_utils as utils
from spike.romancal.datamodels_library import ModelLibrary
from spike.romancal import resample_gwcs_drizzle, resample_utils

log = logging.getLogger(__name__)
log.setLevel(logging.DEBUG)

__all__ = ["OutputTooLargeError", "ResampleData"]


class OutputTooLargeError(MemoryError):
    """Raised when the output is too large for in-memory instantiation"""


class ResampleData:
    """
    This is the controlling routine for the resampling process.

    Notes
    -----
    This routine performs the following operations:

      1. Extracts parameter settings from input model, such as pixfrac,
         weight type, exposure time (if relevant), and kernel, and merges
         them with any user-provided values.
      2. Creates output WCS based on input images and define mapping function
         between all input arrays and the output array. Alternatively, a custom,
         user-provided WCS object can be used instead.
      3. Updates output data model with output arrays from drizzle, including
         a record of metadata from all input models.
    """

    def __init__(
        self,
        input_models,
        output=None,
        single=False,
        blendheaders=True,
        pixfrac=1.0,
        kernel="square",
        fillval="INDEF",
        wht_type="ivm",
        good_bits="0",
        pscale_ratio=1.0,
        pscale=None,
        **kwargs,
    ):
        """
        Parameters
        ----------
        input_models : ~romancal.datamodels.ModelLibrary
            A `~romancal.datamodels.ModelLibrary` object containing the data
            to be processed.

        output : str
            filename for output

        kwargs : dict
            Other parameters.

            .. note::
                ``output_shape`` is in the ``x, y`` order.

            .. note::
                ``in_memory`` controls whether or not the resampled
                array from ``resample_many_to_many()``
                should be kept in memory or written out to disk and
                deleted from memory. Default value is `True` to keep
                all products in memory.
        """
        if (input_models is None) or (len(input_models) == 0):
            raise ValueError(
                "No input has been provided. Input must be a non-empty ModelLibrary"
            )

        self.input_models = input_models
        self.output_filename = output
        self.pscale_ratio = pscale_ratio
        self.single = single
        self.blendheaders = blendheaders
        self.pixfrac = pixfrac
        self.kernel = kernel
        self.fillval = fillval
        self.weight_type = wht_type
        self.good_bits = good_bits
        self.in_memory = kwargs.get("in_memory", True)
        if "target" in input_models.asn:
            self.location_name = input_models.asn["target"]
        else:
            self.location_name = "None"

        log.info(f"Driz parameter kernel: {self.kernel}")
        log.info(f"Driz parameter pixfrac: {self.pixfrac}")
        log.info(f"Driz parameter fillval: {self.fillval}")
        log.info(f"Driz parameter weight_type: {self.weight_type}")

        output_wcs = kwargs.get("output_wcs", None)
        output_shape = kwargs.get("output_shape", None)
        crpix = kwargs.get("crpix", None)
        crval = kwargs.get("crval", None)
        rotation = kwargs.get("rotation", None)

        if pscale is not None:
            log.info(f"Output pixel scale: {pscale} arcsec.")
            pscale /= 3600.0
        else:
            log.info(f"Output pixel scale ratio: {pscale_ratio}")

        # build the output WCS object
        if output_wcs:
            # use the provided WCS object
            self.output_wcs = output_wcs
            if output_shape is not None:
                self.output_wcs.array_shape = output_shape[::-1]
        else:
            with self.input_models:
                models = list(self.input_models)
                # determine output WCS based on all inputs, including a reference WCS
                self.output_wcs = resample_utils.make_output_wcs(
                    models,
                    pscale_ratio=self.pscale_ratio,
                    pscale=pscale,
                    rotation=rotation,
                    shape=None if output_shape is None else output_shape[::-1],
                    crpix=crpix,
                    crval=crval,
                )
                for i, m in enumerate(models):
                    self.input_models.shelve(m, i, modify=False)

        log.debug(f"Output mosaic size: {self.output_wcs.array_shape}")

        # NOTE: should we enable memory allocation?

        # can_allocate, required_memory = datamodels.util.check_memory_allocation(
        #     self.output_wcs.array_shape,
        #     kwargs['allowed_memory'],
        #     datamodels.ImageModel
        # )
        # if not can_allocate:
        #     raise OutputTooLargeError(
        #         f'Combined ImageModel size {self.output_wcs.array_shape} '
        #         f'requires {bytes2human(required_memory)}. '
        #         f'Model cannot be instantiated.'
        #     )

        # NOTE: wait for William to fix bug in datamodels' init and then
        # use datamodels.ImageModel(shape=(nx, ny)) instead of mk_datamodel()

        # n_images sets the number of context image planes.
        # This should be 1 to start (not the default of 2).
        self.blank_output = maker_utils.mk_datamodel(
            datamodels.MosaicModel, n_images=1, shape=tuple(self.output_wcs.array_shape)
        )

        with self.input_models:
            models = list(self.input_models)

            # update meta.basic
            populate_mosaic_basic(self.blank_output, models)

            # update meta.cal_step
            self.blank_output.meta.cal_step = maker_utils.mk_l3_cal_step(
                **models[0].meta.cal_step.to_flat_dict()
            )

            # FIXME: temporary hack to prevent changes from L2 into L3 schema
            for i, model in enumerate(models):
                # saving ImageModels cal_logs
                cal_logs = model.meta.cal_logs
                # removing meta.cal_logs
                del model.meta["cal_logs"]
                # Update the output with all the component metas
                populate_mosaic_individual(self.blank_output, [model])
                # re-attaching cal_logs to meta
                model.meta.cal_logs = cal_logs

            # update meta data and wcs
            l2_into_l3_meta(self.blank_output.meta, models[0].meta)
            self.blank_output.meta.wcs = self.output_wcs
            gwcs_into_l3(self.blank_output, self.output_wcs)

            self.blank_output.cal_logs = stnode.CalLogs()
            self.blank_output["individual_image_cal_logs"] = [
                model.meta.cal_logs for model in models
            ]
            for i, m in enumerate(models):
                self.input_models.shelve(m, i, modify=False)

    def do_drizzle(self):
        """Pick the correct drizzling mode based on ``self.single``."""
        if self.single:
            return self.resample_many_to_many()
        else:
            return self.resample_many_to_one()

    def resample_group(self, input_models, indices):
        """
        Resample models at the provided indices.

        This is used by outlier detection and will not blend metadata
        and will not resample variance, error or exposure time.

        Parameters
        ----------
        input_models : ModelLibrary

        indices : list
            List of model indices to include in this resampling
        """
        output_model = self.blank_output.copy()
        output_model.meta["resample"] = maker_utils.mk_resample()
        output_model.meta.basic.location_name = self.location_name

        copy_asn_info_from_library(input_models, output_model)

        with input_models:
            example_image = input_models.borrow(indices[0])

            # Determine output file type from input exposure filenames
            # Use this for defining the output filename
            indx = example_image.meta.filename.rfind(".")
            output_type = example_image.meta.filename[indx:]
            output_root = "_".join(
                example_image.meta.filename.replace(output_type, "").split("_")[:-1]
            )
            output_model.meta.filename = f"{output_root}_outlier_i2d{output_type}"
            input_models.shelve(example_image, indices[0], modify=False)
            output_model.meta.basic.location_name = self.location_name
            del example_image

            # Initialize the output with the wcs
            driz = gwcs_drizzle.GWCSDrizzle(
                output_model,
                pixfrac=self.pixfrac,
                kernel=self.kernel,
                fillval=self.fillval,
            )

            log.info(f"{len(indices)} exposures to drizzle together")
            for index in indices:
                img = input_models.borrow(index)
                # TODO: should weight_type=None here?
                inwht = resample_utils.build_driz_weight(
                    img, weight_type=self.weight_type, good_bits=self.good_bits
                )

                # apply sky subtraction
                if (
                    hasattr(img.meta, "background")
                    and img.meta.background.subtracted is False
                    and img.meta.background.level is not None
                ):
                    data = img.data - img.meta.background.level
                else:
                    data = img.data

                xmin, xmax, ymin, ymax = resample_utils.resample_range(
                    data.shape, img.meta.wcs.bounding_box
                )

                driz.add_image(
                    data,
                    img.meta.wcs,
                    inwht=inwht,
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )
                del data
                self.input_models.shelve(img, index, modify=False)
                del img

        # cast context array to uint32
        output_model.context = output_model.context.astype("uint32")
        return output_model

    def resample_many_to_many(self):
        """Resample many inputs to many outputs where outputs have a common frame.

        Coadd only different detectors of the same exposure (e.g. map SCA 1 and
        10 onto the same output image), as they image different areas of the
        sky.
        """
        output_models = []
        for group_id, indices in self.input_models.group_indices.items():
            output_model = self.resample_group(self.input_models, indices)

            if not self.in_memory:
                # Write out model to disk, then return filename
                output_name = output_model.meta.filename
                if self.output_dir is not None:
                    output_name = os.path.join(self.output_dir, output_name)
                output_model.save(output_name)
                log.info(f"Saved model in {output_name}")
                output_models.append(output_name)
            else:
                output_models.append(output_model)

        if not self.in_memory:
            # build ModelLibrary as an association from the output files
            # this saves memory if there are multiple groups
            asn = asn_from_list(output_models, product_name="outlier_i2d")
            asn_dict = json.loads(
                asn.dump()[1]
            )  # serializes the asn and converts to dict
            return ModelLibrary(asn_dict, on_disk=True)
        # otherwise just build it as a list of in-memory models
        return ModelLibrary(output_models, on_disk=False)

    def resample_many_to_one(self):
        """Resample and coadd many inputs to a single output.
        Used for level 3 resampling
        """
        output_model = self.blank_output.copy()
        output_model.meta.filename = self.output_filename
        output_model.meta["resample"] = maker_utils.mk_resample()
        output_model.meta.resample.weight_type = self.weight_type
        output_model.meta.resample.pointings = len(self.input_models.group_names)
        output_model.meta.basic.location_name = self.location_name

        # copy over asn information
        if (asn_pool := self.input_models.asn.get("asn_pool", None)) is not None:
            output_model.meta.asn.pool_name = asn_pool
        if (
            asn_table_name := self.input_models.asn.get("table_name", None)
        ) is not None:
            output_model.meta.asn.table_name = asn_table_name

        if self.blendheaders:
            log.info("Skipping blendheaders for now.")

        # Initialize the output with the wcs
        driz = gwcs_drizzle.GWCSDrizzle(
            output_model,
            outwcs=self.output_wcs,
            pixfrac=self.pixfrac,
            kernel=self.kernel,
            fillval=self.fillval,
        )

        log.info("Resampling science data")
        with self.input_models:
            for i, img in enumerate(self.input_models):
                inwht = resample_utils.build_driz_weight(
                    img,
                    weight_type=self.weight_type,
                    good_bits=self.good_bits,
                )
                if (
                    hasattr(img.meta, "background")
                    and img.meta.background.subtracted is False
                    and img.meta.background.level is not None
                ):
                    data = img.data - img.meta.background.level
                else:
                    data = img.data

                xmin, xmax, ymin, ymax = resample_utils.resample_range(
                    data.shape, img.meta.wcs.bounding_box
                )

                driz.add_image(
                    data,
                    img.meta.wcs,
                    inwht=inwht,
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )
                del data, inwht
                self.input_models.shelve(img, i, modify=False)

        # record the actual filenames (the expname from the association)
        # for each file used to generate the output_model
        output_model.meta.resample["members"] = [
            m["expname"] for m in self.input_models.asn["products"][0]["members"]
        ]

        # Resample variances array in self.input_models to output_model
        self.resample_variance_array("var_rnoise", output_model)
        self.resample_variance_array("var_poisson", output_model)
        self.resample_variance_array("var_flat", output_model)

        # Make exposure time image
        exptime_tot = self.resample_exposure_time(output_model)

        # TODO: fix unit here
        output_model.err = np.sqrt(
            np.nansum(
                [
                    output_model.var_rnoise,
                    output_model.var_poisson,
                    output_model.var_flat,
                ],
                axis=0,
            )
        )

        self.update_exposure_times(output_model, exptime_tot)

        # TODO: fix RAD to expect a context image datatype of int32
        output_model.context = output_model.context.astype(np.uint32)

        return ModelLibrary([output_model])

    def resample_variance_array(self, name, output_model):
        """Resample variance arrays from ``self.input_models`` to the ``output_model``.

        Resample the ``name`` variance array to the same name in ``output_model``,
        using a cumulative sum.

        This modifies ``output_model`` in-place.
        """
        output_wcs = self.output_wcs
        inverse_variance_sum = np.full_like(output_model.data, np.nan)

        log.info(f"Resampling {name}")
        with self.input_models:
            for i, model in enumerate(self.input_models):
                variance = getattr(model, name)
                if variance is None or variance.size == 0:
                    log.debug(
                        f"No data for '{name}' for model "
                        f"{repr(model.meta.filename)}. Skipping ..."
                    )
                    continue
                elif variance.shape != model.data.shape:
                    log.warning(
                        f"Data shape mismatch for '{name}' for model "
                        f"{repr(model.meta.filename)}. Skipping..."
                    )
                    continue

                # create a unit weight map for all the input pixels with science data
                inwht = resample_utils.build_driz_weight(
                    model, weight_type=None, good_bits=self.good_bits
                )

                resampled_variance = np.zeros_like(output_model.data)
                outwht = np.zeros_like(output_model.data)
                outcon = np.zeros_like(output_model.context)

                xmin, xmax, ymin, ymax = resample_utils.resample_range(
                    variance.shape, model.meta.wcs.bounding_box
                )

                # resample the variance array (fill "unpopulated" pixels with NaNs)
                self.drizzle_arrays(
                    variance,
                    inwht,
                    model.meta.wcs,
                    output_wcs,
                    resampled_variance,
                    outwht,
                    outcon,
                    pixfrac=self.pixfrac,
                    kernel=self.kernel,
                    fillval=np.nan,
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )

                # Add the inverse of the resampled variance to a running sum.
                # Update only pixels (in the running sum) with valid new values:
                mask = resampled_variance > 0

                inverse_variance_sum[mask] = np.nansum(
                    [
                        inverse_variance_sum[mask],
                        np.reciprocal(resampled_variance[mask]),
                    ],
                    axis=0,
                )
                self.input_models.shelve(model, i, modify=False)

        # We now have a sum of the inverse resampled variances.  We need the
        # inverse of that to get back to units of variance.
        # TODO: fix unit here
        output_variance = np.reciprocal(inverse_variance_sum)

        setattr(output_model, name, output_variance)

    def resample_exposure_time(self, output_model):
        """Resample the exposure time from ``self.input_models`` to the
        ``output_model``.

        Create an exposure time image that is the drizzled sum of the input
        images.
        """
        output_wcs = self.output_wcs
        exptime_tot = np.zeros(output_model.data.shape, dtype="f4")

        log.info("Resampling exposure time")
        with self.input_models:
            for i, model in enumerate(self.input_models):
                exptime = np.full(
                    model.data.shape, model.meta.exposure.effective_exposure_time
                )

                # create a unit weight map for all the input pixels with science data
                inwht = resample_utils.build_driz_weight(
                    model, weight_type=None, good_bits=self.good_bits
                )

                resampled_exptime = np.zeros_like(output_model.data)
                outwht = np.zeros_like(output_model.data)
                outcon = np.zeros_like(output_model.context, dtype="i4")
                # drizzle wants an i4, but datamodels wants a u4.

                xmin, xmax, ymin, ymax = resample_utils.resample_range(
                    exptime.shape, model.meta.wcs.bounding_box
                )

                # resample the exptime array
                self.drizzle_arrays(
                    exptime * u.s,  # drizzle_arrays expects these to have units
                    inwht,
                    model.meta.wcs,
                    output_wcs,
                    resampled_exptime,
                    outwht,
                    outcon,
                    pixfrac=1,  # for exposure time images, always use pixfrac = 1
                    kernel=self.kernel,
                    fillval=0,
                    xmin=xmin,
                    xmax=xmax,
                    ymin=ymin,
                    ymax=ymax,
                )

                exptime_tot += resampled_exptime
                self.input_models.shelve(model, i, modify=False)

        return exptime_tot

    def update_exposure_times(self, output_model, exptime_tot):
        """Update exposure time metadata (in-place)."""
        m = exptime_tot > 0
        total_exposure_time = np.mean(exptime_tot[m]) if np.any(m) else 0
        max_exposure_time = np.max(exptime_tot)
        log.info(
            f"Mean, max exposure times: {total_exposure_time:.1f}, "
            f"{max_exposure_time:.1f}"
        )
        exposure_times = {"start": [], "end": []}
        with self.input_models:
            for group_id, indices in self.input_models.group_indices.items():
                index = indices[0]
                model = self.input_models.borrow(index)
                exposure_times["start"].append(model.meta.exposure.start_time)
                exposure_times["end"].append(model.meta.exposure.end_time)
                self.input_models.shelve(model, index, modify=False)

        # Update some basic exposure time values based on output_model
        output_model.meta.basic.mean_exposure_time = total_exposure_time
        output_model.meta.basic.time_first_mjd = min(exposure_times["start"]).mjd
        output_model.meta.basic.time_last_mjd = max(exposure_times["end"]).mjd
        output_model.meta.basic.max_exposure_time = max_exposure_time
        output_model.meta.resample.product_exposure_time = max_exposure_time

    @staticmethod
    def drizzle_arrays(
        insci,
        inwht,
        input_wcs,
        output_wcs,
        outsci,
        outwht,
        outcon,
        uniqid=1,
        xmin=None,
        xmax=None,
        ymin=None,
        ymax=None,
        pixfrac=1.0,
        kernel="square",
        fillval="INDEF",
        wtscale=1.0,
    ):
        """
        Low level routine for performing 'drizzle' operation on one image.

        The interface is compatible with STScI code. All images are Python
        `ndarrays`, instead of filenames. File handling (input and output) is
        performed by the calling routine.

        Parameters
        ----------

        insci : 2d array
            A 2d `numpy` array containing the input image to be drizzled.

        inwht : 2d array
            A 2d `numpy` array containing the pixel by pixel weighting.
            Must have the same dimensions as insci. If none is supplied,
            the weighting is set to one.

        input_wcs : `gwcs.wcs.WCS` object
            The world coordinate system of the input image.

        output_wcs : `gwcs.wcs.WCS` object
            The world coordinate system of the output image.

        outsci : 2d array
            A 2d `numpy` array containing the output image produced by
            drizzling. On the first call it should be set to zero.
            Subsequent calls it will hold the intermediate results.  This
            is modified in-place.

        outwht : 2d array
            A 2d `numpy` array containing the output counts. On the first
            call it should be set to zero. On subsequent calls it will
            hold the intermediate results.  This is modified in-place.

        outcon : 2d or 3d array, optional
            A 2d or 3d `numpy` array holding a bitmap of which image was an input
            for each output pixel. Should be integer zero on first call.
            Subsequent calls hold intermediate results.  This is modified
            in-place.

        uniqid : int, optional
            The id number of the input image. Should be one the first time
            this function is called and incremented by one on each subsequent
            call.

        xmin : float, None, optional
            on the input image. Only pixels on the input image inside this
            rectangle will have their flux added to the output image. Xmin
            sets the minimum value of the x dimension. The x dimension is the
            dimension that varies quickest on the image. If the value is zero,
            no minimum will be set in the x dimension. All four parameters are
            zero based, counting starts at zero.

        xmax : float, None, optional
            Sets the maximum value of the x dimension on the bounding box
            of the input image. If the value is zero, no maximum will
            be set in the x dimension, the full x dimension of the output
            image is the bounding box.

        ymin : float, None, optional
            Sets the minimum value in the y dimension on the bounding box. The
            y dimension varies less rapidly than the x and represents the line
            index on the input image. If the value is zero, no minimum  will be
            set in the y dimension.

        ymax : float, None, optional
            Sets the maximum value in the y dimension. If the value is zero, no
            maximum will be set in the y dimension,  the full x dimension
            of the output image is the bounding box.

        pixfrac : float, optional
            The fraction of a pixel that the pixel flux is confined to. The
            default value of 1 has the pixel flux evenly spread across the image.
            A value of 0.5 confines it to half a pixel in the linear dimension,
            so the flux is confined to a quarter of the pixel area when the square
            kernel is used.

        kernel: str, optional
            The name of the kernel used to combine the input. The choice of
            kernel controls the distribution of flux over the kernel. The kernel
            names are: `'square'`, `'gaussian'`, `'point'`, `'tophat'`, `'turbo'`,
            `'lanczos2'`, and `'lanczos3'`. The `'square'` kernel is the default.

        fillval: str, optional
            The value a pixel is set to in the output if the input image does
            not overlap it. The default value of INDEF does not set a value.

        Returns
        -------
        : tuple
            A tuple with three values: a version string, the number of pixels
            on the input image that do not overlap the output image, and the
            number of complete lines on the input image that do not overlap the
            output input image.

        """

        # Insure that the fillval parameter gets properly interpreted for use with tdriz
        fillval = "INDEF" if util.is_blank(str(fillval)) else str(fillval)
        if insci.dtype > np.float32:
            insci = insci.astype(np.float32)

        # Add input weight image if it was not passed in
        if inwht is None:
            inwht = np.ones_like(insci)

        # Compute what plane of the context image this input would
        # correspond to:
        planeid = int((uniqid - 1) / 32)

        # Check if the context image has this many planes
        if outcon.ndim == 2:
            nplanes = 1
        elif outcon.ndim == 3:
            nplanes = outcon.shape[0]
        else:
            nplanes = 0

        if nplanes <= planeid:
            raise IndexError("Not enough planes in drizzle context image")

        # Alias context image to the requested plane if 3d
        if outcon.ndim == 3:
            outcon = outcon[planeid]

        # Compute the mapping between the input and output pixel coordinates
        # for use in drizzle.cdrizzle.tdriz
        pixmap = resample_utils.calc_gwcs_pixmap(input_wcs, output_wcs, insci.shape)

        log.debug(f"Pixmap shape: {pixmap[:,:,0].shape}")
        log.debug(f"Input Sci shape: {insci.shape}")
        log.debug(f"Output Sci shape: {outsci.shape}")

        log.info(f"Drizzling {insci.shape} --> {outsci.shape}")

        _vers, _nmiss, _nskip = cdrizzle.tdriz(
            insci.astype(np.float32),
            inwht,
            pixmap,
            outsci,
            outwht,
            outcon,
            uniqid=uniqid,
            xmin=xmin,
            xmax=xmax,
            ymin=ymin,
            ymax=ymax,
            pixfrac=pixfrac,
            kernel=kernel,
            in_units="cps",
            expscale=1.0,
            wtscale=wtscale,
            fillstr=fillval,
        )


def l2_into_l3_meta(l3_meta, l2_meta):
    """Update the level 3 meta with info from the level 2 meta

    Parameters
    ----------
    l3_meta : dict
        The meta to update. This is updated in-place

    l2_meta : stnode
        The Level 2-like meta to pull from

    Notes
    -----
    The list of meta that is pulled from the Level 2 meta into the Level 3 meta is as follows:
    basic.visit: observation.visit
    basic.segment: observation.segment
    basic.pass: observation.pass
    basic.program: observation.program
    basic.optical_element: optical_element
    basic.instrument: instrument.name
    basic.telescope: telescope
    program: program
    """
    l3_meta.basic.visit = l2_meta.observation.visit
    l3_meta.basic.segment = l2_meta.observation.segment
    l3_meta.basic["pass"] = l2_meta.observation["pass"]
    l3_meta.basic.program = l2_meta.observation.program
    l3_meta.basic.optical_element = l2_meta.instrument.optical_element
    l3_meta.basic.instrument = l2_meta.instrument.name
    l3_meta.coordinates = l2_meta.coordinates
    l3_meta.program = l2_meta.program


def gwcs_into_l3(model, wcs):
    """Update the Level 3 wcsinfo block from a GWCS object

    Parameters
    ----------
    model : `DataModel`
        The model whose meta is to be updated.

    wcs : `GWCS`
        GWCS info to transfer into the `meta.wcsinfo` block

    Notes
    -----
    Some models/parameters in the GWCS object have explicit names, such as
    'crpix1'. However, some do not and hence have to be accessed explicitly
    by indexing. This is fragile and will be a source of issues.
    """
    l3_wcsinfo = model.meta.wcsinfo
    transform = wcs.forward_transform

    l3_wcsinfo.projection = "TAN"
    l3_wcsinfo.pixel_shape = model.shape

    pixel_center = [(v - 1) / 2.0 for v in model.shape[::-1]]
    world_center = wcs(*pixel_center)
    l3_wcsinfo.ra_center = world_center[0]
    l3_wcsinfo.dec_center = world_center[1]
    l3_wcsinfo.pixel_scale_local = compute_scale(wcs, world_center)
    l3_wcsinfo.orientat_local = calc_pa(wcs, *world_center)

    try:
        footprint = utils.create_footprint(wcs, model.shape)
    except Exception as excp:
        log.warning("Could not determine footprint due to %s", excp)
    else:
        l3_wcsinfo.ra_corn1 = footprint[0][0]
        l3_wcsinfo.ra_corn2 = footprint[1][0]
        l3_wcsinfo.ra_corn3 = footprint[2][0]
        l3_wcsinfo.ra_corn4 = footprint[3][0]
        l3_wcsinfo.dec_corn1 = footprint[0][1]
        l3_wcsinfo.dec_corn2 = footprint[1][1]
        l3_wcsinfo.dec_corn3 = footprint[2][1]
        l3_wcsinfo.dec_corn4 = footprint[3][1]
        l3_wcsinfo.s_region = compute_s_region_keyword(footprint)

    try:
        l3_wcsinfo.x_ref = -transform["crpix1"].offset.value
        l3_wcsinfo.y_ref = -transform["crpix2"].offset.value
    except IndexError:
        log.warning(
            "WCS has no clear reference pixel defined by crpix1/crpix2. Assuming reference pixel is center."
        )
        l3_wcsinfo.x_ref = pixel_center[0]
        l3_wcsinfo.y_ref = pixel_center[1]
    world_ref = wcs(l3_wcsinfo.x_ref, l3_wcsinfo.y_ref)
    l3_wcsinfo.ra_ref = world_ref[0]
    l3_wcsinfo.dec_ref = world_ref[1]
    l3_wcsinfo.pixel_scale = compute_scale(wcs, world_ref)
    l3_wcsinfo.orientat = calc_pa(wcs, *world_ref)

    try:
        l3_wcsinfo.rotation_matrix = transform[
            "pc_rotation_matrix"
        ].matrix.value.tolist()
    except Exception:
        log.warning(
            "WCS has no clear rotation matrix defined by pc_rotation_matrix. Calculating one."
        )
        rotation_matrix = utils.calc_rotation_matrix(l3_wcsinfo.orientat, 0.0)
        l3_wcsinfo.rotation_matrix = utils.list_1d_to_2d(rotation_matrix, 2)


def calc_pa(wcs, ra, dec):
    """Calculate position angle at given ra,dec

    Parameters
    ----------
    wcs : GWCS
        The wcs in consideration.

    ra, dec : float, float
        The ra/dec in degrees.

    Returns
    -------
    position_angle : float
        The position angle in degrees.

    """
    delta_pix = [v for v in wcs.world_to_pixel(ra, dec)]
    delta_pix[1] += 1
    delta_coord = wcs.pixel_to_world(*delta_pix)
    coord = SkyCoord(ra, dec, frame="icrs", unit="deg")

    return coord.position_angle(delta_coord).degree


def populate_mosaic_basic(
    output_model: datamodels.MosaicModel, input_models: [List, ModelLibrary]
):
    """
    Populate basic metadata fields in the output mosaic model based on input models.

    Parameters
    ----------
    output_model : MosaicModel
        Object to populate with basic metadata.
    input_models : [List, ModelLibrary]
        List of input data models from which to extract the metadata.
        ModelLibrary is also supported.

    Returns
    -------
    None
    """

    input_meta = [datamodel.meta for datamodel in input_models]

    # time data
    output_model.meta.basic.time_first_mjd = np.min(
        [x.exposure.start_time.mjd for x in input_meta]
    )
    output_model.meta.basic.time_last_mjd = np.max(
        [x.exposure.end_time.mjd for x in input_meta]
    )
    output_model.meta.basic.time_mean_mjd = np.mean(
        [x.exposure.mid_time.mjd for x in input_meta]
    )

    # observation data
    output_model.meta.basic.visit = (
        input_meta[0].observation.visit
        if len({x.observation.visit for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic.segment = (
        input_meta[0].observation.segment
        if len({x.observation.segment for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic["pass"] = (
        input_meta[0].observation["pass"]
        if len({x.observation["pass"] for x in input_meta}) == 1
        else -1
    )
    output_model.meta.basic.program = (
        input_meta[0].observation.program
        if len({x.observation.program for x in input_meta}) == 1
        else -1
    )

    # instrument data
    output_model.meta.basic.optical_element = input_meta[0].instrument.optical_element
    output_model.meta.basic.instrument = input_meta[0].instrument.name

    # association product type
    output_model.meta.basic.product_type = "TBD"


def populate_mosaic_individual(
    output_model: datamodels.MosaicModel, input_models: [List, ModelLibrary]
):
    """
    Populate individual meta fields in the output mosaic model based on input models.

    Parameters
    ----------
    output_model : MosaicModel
        Object to populate with basic metadata.
    input_models : [List, ModelLibrary]
        List of input data models from which to extract the metadata.
        ModelLibrary is also supported.

    Returns
    -------
    None
    """

    input_metas = [datamodel.meta for datamodel in input_models]
    for input_meta in input_metas:
        output_model.append_individual_image_meta(input_meta)


def copy_asn_info_from_library(input_models, output_model):
    # copy over asn information
    if (asn_pool := input_models.asn.get("asn_pool", None)) is not None:
        output_model.meta.asn.pool_name = asn_pool
    if (asn_table_name := input_models.asn.get("table_name", None)) is not None:
        output_model.meta.asn.table_name = asn_table_name


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/associations/association.py
"""

from collections.abc import MutableMapping
from datetime import datetime, timezone
import json
import jsonschema
import logging
import re
import os
import warnings

__version__ = "1.16.1" #hardcoding from latest release
from spike.jwstcal.associations_exceptions import (
    AssociationNotValidError
)
from spike.jwstcal.associations_lib_constraint import (
    Constraint,
    meets_conditions
)
from spike.stpipe.format_template import FormatTemplate
from spike.jwstcal.associations_lib_ioregistry import IORegistry

__all__ = ['Association']


# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Timestamp template
_TIMESTAMP_TEMPLATE = '%Y%m%dt%H%M%S'


class Association(MutableMapping):
    """Association Base Class

    Parameters
    ----------
    version_id : str or None
        Version ID to use in the name of this association.
        If None, nothing is added.

    Raises
    ------
    AssociationError
        If an item doesn't match.

    Attributes
    ----------
    instance : dict-like
        The instance is the association data structure.
        See `data` below

    meta : dict
        Information about the association.

    data : dict
        The association. The format of this data structure
        is determined by the individual associations and, if
        defined, validated against their specified schema.

    schema_file : str
        The name of the output schema that an association
        must adhere to.
    """

    registry = None
    """Registry this rule has been placed in."""

    DEFAULT_FORCE_UNIQUE = False
    """Default whether to force constraints to use unique values."""

    DEFAULT_REQUIRE_CONSTRAINT = True
    """Default require that the constraint exists or otherwise
    can be explicitly checked.
    """

    DEFAULT_EVALUATE = False
    """Default do not evaluate input values"""

    GLOBAL_CONSTRAINT = None
    """Global constraints"""

    INVALID_VALUES: tuple | None = None
    """Attribute values that indicate the
    attribute is not specified.
    """

    ioregistry: IORegistry = IORegistry()
    """The association IO registry"""

    def __init__(
            self,
            version_id=None,
    ):

        self.data = dict()
        self.run_init_hook = True
        self.meta = {}

        self.version_id = version_id

        self.data.update({
            'asn_type': 'None',
            'asn_rule': self.asn_rule,
            'version_id': self.version_id,
            'code_version': __version__,
        })

        # Setup constraints
        # These may be predefined by a rule.
        try:
            constraints = self.constraints
        except AttributeError:
            constraints = Constraint()
        if self.GLOBAL_CONSTRAINT is not None:
            constraints.append(self.GLOBAL_CONSTRAINT.copy())
        self.constraints = constraints

    @classmethod
    def create(cls, item, version_id=None):
        """Create association if item belongs

        Parameters
        ----------
        item : dict
            The item to initialize the association with.

        version_id : str or None
            Version ID to use in the name of this association.
            If None, nothing is added.

        Returns
        -------
        (association, reprocess_list)
            2-tuple consisting of:
                - association or None: The association or, if the item does not
                  match this rule, None
                - [ProcessList[, ...]]: List of items to process again.
        """
        asn = cls(version_id=version_id)
        matches, reprocess = asn.add(item)
        if not matches:
            return None, reprocess
        return asn, reprocess

    @property
    def asn_name(self):
        """Suggest filename for the association"""
        return 'unnamed_association'

    @classmethod
    def _asn_rule(cls):
        return cls.__name__

    @property
    def asn_rule(self):
        """Name of the rule"""
        return self._asn_rule()

    @classmethod
    def validate(cls, asn):
        """Validate an association against this rule

        Parameters
        ----------
        asn : Association or association-like
            The association structure to examine

        Returns
        -------
        valid : bool
            True if valid. Otherwise the `AssociationNotValidError` is raised

        Raises
        ------
        AssociationNotValidError
            If there is some reason validation failed.

        Notes
        -----
        The base method checks against the rule class' schema
        If the rule class does not define a schema, a warning is issued
        but the routine will return True.
        """
        if not hasattr(cls, 'schema_file'):
            logger.warning(
                'Cannot validate: {} has no schema. Presuming OK.'.format(cls)
            )
            return True

        if isinstance(asn, cls):
            asn_data = asn.data
        else:
            asn_data = asn

        with open(cls.schema_file, 'r') as schema_file:
            asn_schema = json.load(schema_file)

        try:
            jsonschema.validate(asn_data, asn_schema)
        except (AttributeError, jsonschema.ValidationError) as err:
            logger.debug('Validation failed:')
            logger.debug(str(err))
            raise AssociationNotValidError('Validation failed') from err

        # Validate no path data for expnames
        for product in asn_data["products"]:
            members = product['members']
            for member in members:
                fpath, fname = os.path.split(member["expname"])
                if len(fpath) > 0:
                    err_str = "Input association file contains path information;"
                    err_str += " note that this can complicate usage and/or sharing"
                    err_str += " of such files."
                    logger.debug(err_str)
                    warnings.warn(err_str, UserWarning)
        return True

    def dump(self, format='json', **kwargs):
        """Serialize the association

        Parameters
        ----------
        format : str
            The format to use to dump the association into.

        kwargs : dict
            List of arguments to pass to the registered
            routines for the current association type.

        Returns
        -------
        (name, serialized):
            Tuple where the first item is the suggested
            base name for the file.
            Second item is the serialization.

        Raises
        ------
        AssociationError
            If the operation cannot be done

        AssociationNotValidError
            If the given association does not validate.
        """
        if self.is_valid:
            return self.ioregistry[format].dump(self, **kwargs)
        else:
            raise AssociationNotValidError(
                'Association {} is not valid'.format(self)
            )

    @classmethod
    def load(
            cls,
            serialized,
            format=None,
            validate=True,
            **kwargs
    ):
        """Marshall a previously serialized association

        Parameters
        ----------
        serialized : object
            The serialized form of the association.

        format : str or None
            The format to force. If None, try all available.

        validate : bool
            Validate against the class' defined schema, if any.

        kwargs : dict
            Other arguments to pass to the `load` method

        Returns
        -------
        association : Association
            The association.

        Raises
        ------
        AssociationNotValidError
            Cannot create or validate the association.

        Notes
        -----
        The `serialized` object can be in any format
        supported by the registered I/O routines. For example, for
        `json` and `yaml` formats, the input can be either a string or
        a file object containing the string.
        """
        if format is None:
            formats = [
                format_func
                for format_name, format_func in cls.ioregistry.items()
            ]
        else:
            formats = [cls.ioregistry[format]]

        for format_func in formats:
            try:
                asn = format_func.load(
                    cls, serialized, **kwargs
                )
            except AssociationNotValidError:
                continue
            else:
                break
        else:
            raise AssociationNotValidError(
                'Cannot translate "{}" to an association'.format(serialized)
            )

        # Validate
        if validate:
            cls.validate(asn)

        return asn

    @property
    def is_valid(self):
        """Check if association is valid"""
        try:
            self.__class__.validate(self)
        except AssociationNotValidError:
            return False
        return True

    def add(self, item, check_constraints=True):
        """Add the item to the association

        Parameters
        ----------
        item : dict
            The item to add.

        check_constraints : bool
            If True, see if the item should belong to this association.
            If False, just add it.

        Returns
        -------
        (match, reprocess_list)
            2-tuple consisting of:
                - bool : True if match
                - [ProcessList[, ...]]: List of items to process again.
        """
        if self.is_item_member(item):
            return True, []

        match = not check_constraints
        if check_constraints:
            match, reprocess = self.check_and_set_constraints(item)

        if match:
            if self.run_init_hook:
                self._init_hook(item)
                self.run_init_hook = False
            self._add(item)

        # If a constraint `force_match` exists, set the `match`
        # result to the value of the constraint.
        try:
            force_match = self.constraints['force_match'].value
        except (KeyError, TypeError):
            pass
        else:
            if force_match is not None:
                match = force_match

        return match, reprocess

    def check_and_set_constraints(self, item):
        """Check whether the given dictionaries match parameters for
        for this association

        Parameters
        ----------
        item : dict
            The parameters to check/set for this association.
            This can be a list of dictionaries.

        Returns
        -------
        (match, reprocess)
            2-tuple consisting of:
                - bool : Did constraint match?
                - [ProcessItem[, ...]]: List of items to process again.

        """
        self.constraints.preserve()
        match, reprocess = self.constraints.check_and_set(item)
        if not match:
            self.constraints.restore()

        # Set the association type for all reprocessed items.
        for process_list in reprocess:
            process_list.trigger_rules.update([type(self)])
            if process_list.rules is None:
                process_list.rules = [type(self)]

        return match, reprocess

    def match_constraint(self, item, constraint, conditions):
        """Generic constraint checking

        Parameters
        ----------
        item : dict
            The item to retrieve the values from

        constraint : str
            The name of the constraint

        conditions : dict
            The conditions structure

        Returns
        -------
        (matches, reprocess_list)
            2-tuple consisting of:
                - bool : True if the all constraints are satisfied
                - [ProcessList[, ...]]: List of items to process again.
        """
        reprocess = []
        evaled_str = conditions['inputs'](item)
        if conditions['value'] is not None:
            if not meets_conditions(
                    evaled_str, conditions['value']
            ):
                return False, reprocess

        # At this point, the constraint has passed.
        # Fix the conditions.
        escaped_value = re.escape(evaled_str)
        conditions['found_values'].add(escaped_value)
        if conditions['value'] is None or \
           conditions.get('force_unique', self.DEFAULT_FORCE_UNIQUE):
            conditions['value'] = escaped_value
            conditions['force_unique'] = False

        # That's all folks
        return True, reprocess

    def finalize(self):
        """Finalize association

        Finalize or close-off this association. Perform validations,
        modifications, etc. to ensure that the association is
        complete.

        Returns
        -------
        associations : [association[, ...]] or None
            List of fully-qualified associations that this association
            represents.
            `None` if a complete association cannot be produced.

        """
        if self.is_valid:
            return [self]
        else:
            return None

    def is_item_member(self, item):
        """Check if item is already a member of this association

        Parameters
        ----------
        item : dict
            The item to add.

        Returns
        -------
        is_item_member : bool
            True if item is a member.
        """
        raise NotImplementedError(
            'Association.is_item_member must be implemented by a specific association rule.'
        )

    def _init_hook(self, item):
        """Post-check and pre-item-adding initialization."""
        pass

    def _add(self, item):
        """Add a item, association-specific"""
        raise NotImplementedError(
            'Association._add must be implemented by a specific association rule.'
        )

    def _add_items(self, items, **kwargs):
        """ Force adding items to the association

        Parameters
        ----------
        items : [object[, ...]]
            A list of items to make members of the association.

        Notes
        -----
        This is a low-level shortcut into adding members, such as file names,
        to an association. All defined shortcuts and other initializations are
        by-passed, resulting in a potentially unusable association.
        """
        try:
            self['members'].update(items)
        except KeyError:
            self['members'] = items

    # #################################################
    # Methods required for implementing MutableMapping
    # #################################################
    def __getitem__(self, key):
        return self.data[self.__keytransform__(key)]

    def __setitem__(self, key, value):
        self.data[self.__keytransform__(key)] = value

    def __delitem__(self, key):
        del self.data[self.__keytransform__(key)]

    def __iter__(self):
        return iter(self.data)

    def __len__(self):
        return len(self.data)

    def __keytransform__(self, key):
        return key

    def keys(self):
        return self.data.keys()

    def items(self):
        return self.data.items()

    def values(self):
        return self.data.values()


# #########
# Utilities
# #########
def finalize(asns):
    """Finalize associations by calling their `finalize_hook` method

    Notes
    -----
    This is a functioning example of a finalize callback, and can be used
    as the generic callback. Suggested usage is as follows:

    .. code-block:: python

       from jwst.associations.association import finalize as generic_finalize
       RegistryMarker.callback('finalize')(generic_finalize)
    """
    finalized_asns = list(filter(
        lambda asn: asn is not None,
        map(lambda asn: asn.finalize(), asns)
    ))
    return finalized_asns


def make_timestamp():
    timestamp = datetime.now(timezone.utc).strftime(
        _TIMESTAMP_TEMPLATE
    )
    return timestamp


# Define default product name filling
format_product = FormatTemplate()


"""
This module is designed to make stpipe functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


stpipe copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/stpipe/blob/main/src/stpipe/step.py
"""

import gc
import os
import sys
import warnings
from collections.abc import Sequence
from contextlib import contextmanager, suppress
from functools import partial
from os.path import (
    abspath,
    basename,
    dirname,
    expanduser,
    expandvars,
    isfile,
    join,
    split,
    splitext,
)
from pathlib import Path
from typing import ClassVar

import yaml

try:
    from astropy.io import fits

    DISCOURAGED_TYPES = (fits.HDUList,)
except ImportError:
    DISCOURAGED_TYPES = None

from spike.stpipe import config, config_parser, crds_client, log, utilities
from spike.stpipe.datamodel import AbstractDataModel
from spike.stpipe.format_template import FormatTemplate
from spike.stpipe.library import AbstractModelLibrary
from spike.stpipe.utilities import _not_set


class Step:
    """
    Step
    """

    spec = """
    pre_hooks          = list(default=list())        # List of Step classes to run before step
    post_hooks         = list(default=list())        # List of Step classes to run after step
    output_file        = output_file(default=None)   # File to save output to.
    output_dir         = string(default=None)        # Directory path for output files
    output_ext         = string()                    # Default type of output
    output_use_model   = boolean(default=False)      # When saving use `DataModel.meta.filename`
    output_use_index   = boolean(default=True)       # Append index.
    save_results       = boolean(default=False)      # Force save results
    skip               = boolean(default=False)      # Skip this step
    suffix             = string(default=None)        # Default suffix for output files
    search_output_file = boolean(default=True)       # Use outputfile define in parent step
    input_dir          = string(default=None)        # Input directory
    """  # noqa: E501
    # Nickname used to refer to this class in lieu of the fully-qualified class
    # name.  Must be globally unique!
    class_alias = None

    # Correction parameters. These store and use whatever information a Step
    # may need to perform its operations without re-calculating, or to use
    # from a previous run of the Step.  The structure is up to each Step.
    correction_pars = None
    use_correction_pars = False

    # Reference types for both command line override
    # definition and reference prefetch
    reference_file_types: ClassVar = []

    # Set to False in subclasses to skip prefetch,
    # but by default attempt to prefetch
    prefetch_references = True

    # This needs to be set to a logging formatter for any
    # log_records to be saved.
    _log_records_formatter = None

    @classmethod
    def get_config_reftype(cls):
        """
        Get the CRDS reftype for this step's config reference.

        Returns
        -------
        str
        """
        return f"pars-{cls.__name__.lower()}"

    @classmethod
    def merge_config(cls, config, config_file):
        return config

    @classmethod
    def load_spec_file(cls, preserve_comments=_not_set):
        spec = config_parser.get_merged_spec_file(
            cls, preserve_comments=preserve_comments
        )
        # Add arguments for all of the expected reference files
        for reference_file_type in cls.reference_file_types:
            override_name = crds_client.get_override_name(reference_file_type)
            spec[override_name] = "is_string_or_datamodel(default=None)"
            spec.inline_comments[override_name] = (
                f"# Override the {reference_file_type} reference file"
            )
        return spec

    @classmethod
    def print_configspec(cls):
        specfile = cls.load_spec_file()
        specfile.write(sys.stdout.buffer)

    @classmethod
    def from_config_file(cls, config_file, parent=None, name=None):
        """
        Create a step from a configuration file.

        Parameters
        ----------
        config_file : path or readable file-like object
            The config file to load parameters from

        parent : Step instance, optional
            The parent step of this step.  Used to determine a
            fully-qualified name for this step, and to determine
            the mode in which to run this step.

        name : str, optional
            If provided, use that name for the returned instance.
            If not provided, the following are tried (in order):
            - The ``name`` parameter in the config file
            - The filename of the config file
            - The name of returned class

        Returns
        -------
        step : Step instance
            If the config file has a ``class`` parameter, the return
            value will be as instance of that class.  The ``class``
            parameter in the config file must specify a subclass of
            ``cls``.  If the configuration file has no ``class``
            parameter, then an instance of ``cls`` is returned.

            Any parameters found in the config file will be set
            as member variables on the returned `Step` instance.
        """
        config = config_parser.load_config_file(config_file)

        # If a file object was passed in, pass the file name along
        if hasattr(config_file, "name"):
            config_file = config_file.name

        step_class, name = cls._parse_class_and_name(
            config,
            parent,
            name,
            config_file,
        )

        return step_class.from_config_section(
            config,
            parent=parent,
            name=name,
            config_file=config_file,
        )

    @staticmethod
    def from_cmdline(args):
        """
        Create a step from a configuration file.

        Parameters
        ----------
        args : list of str
            Commandline arguments

        Returns
        -------
        step : Step instance
            If the config file has a ``class`` parameter, the return
            value will be as instance of that class.

            Any parameters found in the config file will be set
            as member variables on the returned `Step` instance.
        """
        from . import cmdline

        return cmdline.step_from_cmdline(args)

    @classmethod
    def _parse_class_and_name(
        cls,
        config,
        parent=None,
        name=None,
        config_file=None,
    ):
        if "class" in config:
            step_class = utilities.import_class(
                utilities.resolve_step_class_alias(config["class"]),
                config_file=config_file,
            )
            if not issubclass(step_class, cls):
                raise TypeError(
                    "Configuration file does not match the expected step class. "
                    f" Expected {cls}, got {step_class}"
                )
        else:
            step_class = cls

        if not name:
            name = config.get("name")
            if not name:
                if isinstance(config_file, str):
                    name = splitext(basename(config_file))[0]
                else:
                    name = step_class.__name__

        if "name" in config:
            del config["name"]
        if "class" in config:
            del config["class"]

        return step_class, name

    @classmethod
    def from_config_section(
        cls,
        config,
        parent=None,
        name=None,
        config_file=None,
    ):
        """
        Create a step from a configuration file fragment.

        Parameters
        ----------
        config : configobj.Section instance
            The config file fragment containing parameters for this
            step only.
        parent : Step instance, optional
            The parent step of this step.  Used to determine a
            fully-qualified name for this step, and to determine
            the mode in which to run this step.
        name : str, optional
            If provided, use that name for the returned instance.
            If not provided, try the following (in order):
            - The ``name`` parameter in the config file fragment
            - The name of returned class
        config_file : str or pathlib.Path, optional
            The path to the config file that created this step, if
            any.  This is used to resolve relative file name
            parameters in the config file.

        Returns
        -------
        step : instance of cls
            Any parameters found in the config file fragment will be
            set as member variables on the returned `Step` instance.
        """
        if not name:
            if config.get("name"):
                name = config["name"]
            else:
                name = cls.__name__

        if "name" in config:
            del config["name"]
        if "class" in config:
            del config["class"]
        if "config_file" in config:
            del config["config_file"]

        spec = cls.load_spec_file()
        config = cls.merge_config(config, config_file)
        config_parser.validate(config, spec, root_dir=dirname(config_file or ""))

        if "config_file" in config:
            del config["config_file"]
        if "name" in config:
            del config["name"]

        # cmdline.FromCommandLine instances should not be passed to
        # steps. Instead, convert them back to strings.
        from . import cmdline

        kwargs = {}
        for k in config:
            if isinstance(config[k], cmdline.FromCommandLine):
                kwargs[k] = str(config[k])
            else:
                kwargs[k] = config[k]

        return cls(
            name=name,
            parent=parent,
            config_file=config_file,
            _validate_kwds=False,
            **kwargs,
        )

    def __init__(
        self,
        name=None,
        parent=None,
        config_file=None,
        _validate_kwds=True,
        **kws,
    ):
        """
        Create a `Step` instance.

        Parameters
        ----------
        name : str, optional
            The name of the Step instance.  Used in logging messages
            and in cache filenames.  If not provided, one will be
            generated based on the class name.

        parent : Step instance, optional
            The parent step of this step.  Used to determine a
            fully-qualified name for this step, and to determine
            the mode in which to run this step.

        config_file : str or pathlib.Path, optional
            The path to the config file that this step was initialized
            with.  Use to determine relative path names of other config files.

        **kws : dict
            Additional parameters to set.  These will be set as member
            variables on the new Step instance.
        """
        self._reference_files_used = []
        # A list of logging.LogRecord emitted to the stpipe root logger
        # during the most recent call to Step.run.
        self._log_records = []
        self._input_filename = None
        self._input_dir = None
        self._keywords = kws
        if _validate_kwds:
            spec = self.load_spec_file()
            kws = config_parser.config_from_dict(
                kws,
                spec,
                root_dir=dirname(config_file or ""),
            )

        if name is None:
            name = self.__class__.__name__
        self.name = name
        if parent is None:
            self.qualified_name = f"{log.STPIPE_ROOT_LOGGER}.{self.name}"
        else:
            self.qualified_name = f"{parent.qualified_name}.{self.name}"
        self.parent = parent

        # Set the parameters as member variables
        for key, val in kws.items():
            setattr(self, key, val)

        # Create a new logger for this step
        self.log = log.getLogger(self.qualified_name)

        self.log.setLevel(log.logging.DEBUG)

        # Log the fact that we have been init-ed.
        self.log.info(
            "%s instance created.",
            self.__class__.__name__,
        )

        # Store the config file path so config filenames can be resolved
        # against it.
        self.config_file = config_file

        # Setup the hooks
        if len(self.pre_hooks) or len(self.post_hooks):
            from . import hooks

            self._pre_hooks = hooks.get_hook_objects(self, "pre", self.pre_hooks)
            self._post_hooks = hooks.get_hook_objects(self, "post", self.post_hooks)
        else:
            self._pre_hooks = []
            self._post_hooks = []

    def _check_args(self, args, discouraged_types, msg):
        if discouraged_types is None:
            return

        if type(args) not in (list, tuple):
            args = [args]

        for i, arg in enumerate(args):
            if isinstance(arg, discouraged_types):
                self.log.error(
                    "%s %s object.  Use an instance of AbstractDataModel instead.",
                    msg,
                    i,
                )

    @property
    def log_records(self):
        """
        Retrieve logs from the most recent run of this step.

        Returns
        -------
        list of str
        """
        return self._log_records

    def run(self, *args):
        """
        Run handles the generic setup and teardown that happens with
        the running of each step.  The real work that is unique to
        each step type is done in the `process` method.
        """
        gc.collect()

        with log.record_logs(formatter=self._log_records_formatter) as log_records:
            self._log_records = log_records

            # Make generic log messages go to this step's logger
            orig_log = log.delegator.log
            log.delegator.log = self.log

            step_result = None

            self.log.info("Step %s running with args %s.", self.name, args)
            # log Step or Pipeline parameters from top level only
            if self.parent is None:
                self.log.info(
                    "Step %s parameters are:%s",
                    self.name,
                    # Add an indent to each line of the YAML output
                    "\n  "
                    + "\n  ".join(
                        yaml.dump(self.get_pars(), sort_keys=False)
                        .strip()
                        # Convert serialized YAML types true/false/null to Python types
                        .replace(" false", " False")
                        .replace(" true", " True")
                        .replace(" null", " None")
                        .splitlines()
                    ),
                )

            if len(args):
                self.set_primary_input(args[0])

            try:
                # Default output file configuration
                if self.output_file is not None:
                    self.save_results = True

                if self.suffix is None:
                    self.suffix = self.default_suffix()

                hook_args = args
                for pre_hook in self._pre_hooks:
                    hook_results = pre_hook.run(*hook_args)
                    if hook_results is not None:
                        hook_args = hook_results
                args = hook_args

                self._reference_files_used = []

                # Warn if passing in objects that should be
                # discouraged.
                self._check_args(args, DISCOURAGED_TYPES, "Passed")

                # Run the Step-specific code.
                if self.skip:
                    self.log.info("Step skipped.")
                    if isinstance(args[0], AbstractModelLibrary):
                        library = args[0]
                        with library:
                            for i, model in enumerate(library):
                                try:
                                    setattr(
                                        model.meta.cal_step, self.class_alias, "SKIPPED"
                                    )
                                except AttributeError as e:
                                    self.log.info(
                                        "Could not record skip into DataModel "
                                        "header: %s",
                                        e,
                                    )
                                library.shelve(model, i)
                    elif isinstance(args[0], AbstractDataModel):
                        if self.class_alias is not None:
                            if isinstance(args[0], Sequence):
                                for model in args[0]:
                                    try:
                                        model[f"meta.cal_step.{self.class_alias}"] = (
                                            "SKIPPED"
                                        )
                                    except AttributeError as e:  # noqa: PERF203
                                        self.log.info(
                                            "Could not record skip into DataModel "
                                            "header: %s",
                                            e,
                                        )
                            elif isinstance(args[0], AbstractDataModel):
                                try:
                                    args[0][
                                        f"meta.cal_step.{self.class_alias}"
                                    ] = "SKIPPED"
                                except AttributeError as e:
                                    self.log.info(
                                        "Could not record skip into DataModel"
                                        " header: %s",
                                        e,
                                    )
                    step_result = args[0]
                else:
                    if self.prefetch_references:
                        self.prefetch(*args)
                    try:
                        step_result = self.process(*args)
                    except TypeError as e:
                        if "process() takes exactly" in str(e):
                            raise TypeError(
                                "Incorrect number of arguments to step"
                            ) from e
                        raise

                # Warn if returning a discouraged object
                self._check_args(step_result, DISCOURAGED_TYPES, "Returned")

                # Run the post hooks
                for post_hook in self._post_hooks:
                    hook_results = post_hook.run(step_result)
                    if hook_results is not None:
                        step_result = hook_results

                # Update meta information
                if isinstance(step_result, AbstractModelLibrary):
                    step_result.finalize_result(self, self._reference_files_used)
                else:
                    if not isinstance(step_result, Sequence):
                        results = [step_result]
                    else:
                        results = step_result

                    # The finalize_result hook allows subclasses to add
                    # metadata (like the cal code package version) before
                    # the result is saved.
                    for result in results:
                        self.finalize_result(result, self._reference_files_used)

                self._reference_files_used = []

                # Save the output file if one was specified
                if not self.skip and self.save_results:
                    # Setup the save list.
                    if not isinstance(step_result, list | tuple):
                        results_to_save = [step_result]
                    else:
                        results_to_save = step_result

                    for idx, result in enumerate(results_to_save):
                        if len(results_to_save) <= 1:
                            idx = None
                        if isinstance(
                            result, (AbstractDataModel | AbstractModelLibrary)
                        ):
                            self.save_model(result, idx=idx)
                        elif hasattr(result, "save"):
                            try:
                                output_path = self.make_output_path(idx=idx)
                            except AttributeError:
                                self.log.warning(
                                    "`save_results` has been requested, but cannot"
                                    " determine filename."
                                )
                                self.log.warning(
                                    "Specify an output file with `--output_file` or set"
                                    " `--save_results=false`"
                                )
                            else:
                                self.log.info("Saving file %s", output_path)
                                result.save(output_path, overwrite=True)

                if not self.skip:
                    self.log.info("Step %s done", self.name)
            finally:
                log.delegator.log = orig_log

        return step_result

    def __call__(self, *args):
        warnings.warn(
            "Step.__call__ is deprecated. It is equivalent to Step.run "
            "and is not recommended.",
            UserWarning,
        )
        return self.run(*args)

    def finalize_result(self, result, reference_files_used):
        """
        Hook that allows subclasses to set mission-specific metadata on each
        step result before that result is saved.

        Parameters
        ----------
        result : a datamodel that is an instance of AbstractDataModel or
                 collections.abc.Sequence
                 One step result (potentially of many).

        reference_files_used : list of tuple
            List of reference files used when running the step, each
            a tuple in the form (str reference type, str reference URI).
        """

    @staticmethod
    def remove_suffix(name):
        """
        Remove a known Step filename suffix from a filename
        (if present).

        Parameters
        ----------
        name : str
            Filename.

        Returns
        -------
        str
            Filename with any known suffix removed.
        str
            Separator that delimited the original suffix.
        """
        return name, "_"

    def prefetch(self, *args):
        """Prefetch reference files,  nominally called when
        self.prefetch_references is True.  Can be called explicitly
        when self.prefetch_refences is False.
        """
        # prefetch truly occurs at the Pipeline (or subclass) level.
        if len(args) and len(self.reference_file_types) and not self.skip:
            self._precache_references(args[0])

    def process(self, *args):
        """
        This is where real work happens. Every Step subclass has to
        override this method. The default behaviour is to raise a
        NotImplementedError exception.
        """
        raise NotImplementedError("Steps have to override process().")

    @classmethod
    def call(cls, *args, **kwargs):
        """
        Creates and runs a new instance of the class.

        Gets a config file from CRDS if one is available

        To set configuration parameters, pass a ``config_file`` path or
        keyword arguments.  Keyword arguments override those in the
        specified ``config_file``.

        Any positional ``*args`` will be passed along to the step's
        ``process`` method.

        Note: this method creates a new instance of `Step` with the given
        ``config_file`` if supplied, plus any extra ``*args`` and ``**kwargs``.
        If you create an instance of a Step, set parameters, and then use
        this ``call()`` method, it will ignore previously-set parameters, as
        it creates a new instance of the class with only the ``config_file``,
        ``*args`` and ``**kwargs`` passed to the ``call()`` method.

        If not used with a ``config_file`` or specific ``*args`` and ``**kwargs``,
        it would be better to use the `run` method, which does not create
        a new instance but simply runs the existing instance of the `Step`
        class.
        """
        filename = None
        if len(args) > 0:
            filename = args[0]
        config, config_file = cls.build_config(filename, **kwargs)

        if "class" in config:
            del config["class"]

        if "logcfg" in config:
            try:
                log.load_configuration(config["logcfg"])
            except Exception as e:
                raise RuntimeError(
                    f"Error parsing logging config {config['logcfg']}"
                ) from e
            del config["logcfg"]

        name = config.get("name", None)
        instance = cls.from_config_section(config, name=name, config_file=config_file)

        return instance.run(*args)

    @property
    def input_dir(self):
        return self.search_attr("_input_dir", "")

    @input_dir.setter
    def input_dir(self, input_dir):
        self._input_dir = input_dir

    def default_output_file(self, input_file=None):
        """Create a default filename based on the input name"""
        output_file = input_file
        if output_file is None or not isinstance(output_file, str):
            output_file = self.search_attr("_input_filename")
        if output_file is None:
            output_file = f"step_{self.name}{self.output_ext}"
        return output_file

    def default_suffix(self):
        """Return a default suffix based on the step"""
        return self.name.lower()

    def search_attr(self, attribute, default=None, parent_first=False):
        """Return first non-None attribute in step hierarchy

        Parameters
        ----------
        attribute : str
            The attribute to retrieve

        default : obj
            If attribute is not found, the value to use

        parent_first : bool
            If `True`, allow parent definition to override step version

        Returns
        -------
        value : obj
            Attribute value or default if not found
        """
        if parent_first:
            try:
                value = self.parent.search_attr(attribute, parent_first=parent_first)
            except AttributeError:
                value = None
            if value is None:
                value = getattr(self, attribute, default)
            return value

        value = getattr(self, attribute, None)
        if value is None:
            try:
                value = self.parent.search_attr(attribute)
            except AttributeError:
                pass
        if value is None:
            value = default
        return value

    def _precache_references(self, input_file):
        """Because Step precaching precedes calls to get_reference_file() almost
        immediately, true precaching has been moved to Pipeline where the
        interleaving of precaching and Step processing is more of an
        issue. This null method is intended to be overridden in Pipeline by
        true precache operations and avoids having to override the more complex
        Step.run() instead.
        """

    def get_ref_override(self, reference_file_type):
        """Determine and return any override for ``reference_file_type``.

        Returns
        -------
        override_filepath or None.
        """
        override_name = crds_client.get_override_name(reference_file_type)
        path = getattr(self, override_name, None)
        if isinstance(path, AbstractDataModel):
            return path

        return abspath(path) if path and path != "N/A" else path

    def get_reference_file(self, input_file, reference_file_type):
        """
        Get a reference file from CRDS.

        If the configuration file or commandline parameters override the
        reference file, it will be automatically used when calling this
        function.

        Parameters
        ----------
        input_file : a datamodel that is an instance of AbstractDataModel
            A model of the input file.  Metadata on this input file
            will be used by the CRDS "bestref" algorithm to obtain a
            reference file.

        reference_file_type : string
            The type of reference file to retrieve.  For example, to
            retrieve a flat field reference file, this would be 'flat'.

        Returns
        -------
        reference_file : path of reference file,  a string
        """
        override = self.get_ref_override(reference_file_type)
        if override is not None:
            if isinstance(override, AbstractDataModel):
                self._reference_files_used.append(
                    (reference_file_type, override.override_handle)
                )
                return override

            if override.strip() != "":
                self._reference_files_used.append(
                    (reference_file_type, basename(override))
                )
                reference_name = override
            else:
                return ""
        else:
            with self.open_model(input_file) as model:
                reference_name = crds_client.get_reference_file(
                    model.get_crds_parameters(),
                    reference_file_type,
                    model.crds_observatory,
                )
            if reference_name != "N/A":
                hdr_name = "crds://" + basename(reference_name)
            else:
                hdr_name = "N/A"
            self._reference_files_used.append((reference_file_type, hdr_name))
        return crds_client.check_reference_open(reference_name)

    @classmethod
    def get_config_from_reference(cls, dataset, disable=None, crds_observatory=None):
        """Retrieve step parameters from reference database

        Parameters
        ----------
        cls : stpipe.Step
            Either a class or instance of a class derived
            from `Step`.
        dataset : A datamodel that is an instance of AbstractDataModel
            A model of the input file.  Metadata on this input file will
            be used by the CRDS "bestref" algorithm to obtain a reference
            file.
        disable: bool or None
            Do not retrieve parameters from CRDS. If None, check global settings.
        crds_observatory : str
            Observatory name ('jwst' or 'roman').

        Returns
        -------
        step_parameters : configobj
            The parameters as retrieved from CRDS. If there is an issue, log as such
            and return an empty config obj.
        """

        # Get the root logger, since the following operations
        # are happening in the surrounding architecture.
        logger = log.delegator.log
        reftype = cls.get_config_reftype()

        if isinstance(dataset, dict):
            # crds_parameters was passed as input from pipeline.py
            crds_parameters = dataset
            crds_observatory = crds_observatory
            if crds_observatory is None:
                raise ValueError("Need a valid name for crds_observatory.")
        else:
            # If the dataset is not an operable instance of AbstractDataModel,
            # log as such and return an empty config object
            try:
                with cls._datamodels_open(dataset, asn_n_members=1) as model:
                    if isinstance(model, Sequence):
                        # Pull out first model in ModelContainer
                        model = model[0]
                    crds_parameters = model.get_crds_parameters()
                    crds_observatory = model.crds_observatory
            except (OSError, TypeError, ValueError):
                logger.warning("Input dataset is not an instance of AbstractDataModel.")
                disable = True

        # Check if retrieval should be attempted.
        if disable is None:
            disable = get_disable_crds_steppars()
        if disable:
            logger.info(
                "%s: CRDS parameter reference retrieval disabled.", reftype.upper()
            )
            return config_parser.ConfigObj()

        # Retrieve step parameters from CRDS
        logger.debug("Retrieving step %s parameters from CRDS", reftype.upper())
        try:
            ref_file = crds_client.get_reference_file(
                crds_parameters,
                reftype,
                crds_observatory,
            )
        except (AttributeError, crds_client.CrdsError):
            logger.debug("%s: No parameters found", reftype.upper())
            return config_parser.ConfigObj()
        if ref_file != "N/A":
            logger.info("%s parameters found: %s", reftype.upper(), ref_file)
            ref = config_parser.load_config_file(ref_file)

            ref_pars = {
                par: value for par, value in ref.items() if par not in ["class", "name"]
            }
            logger.debug(
                "%s parameters retrieved from CRDS: %s", reftype.upper(), ref_pars
            )

            return ref

        logger.debug("No %s reference files found.", reftype.upper())
        return config_parser.ConfigObj()

    def set_primary_input(self, obj, exclusive=True):
        """
        Sets the name of the master input file and input directory.
        Used to generate output file names.

        Parameters
        ----------
        obj : str, pathlib.Path, or instance of AbstractDataModel
            The object to base the name on. If a datamodel,
            use Datamodel.meta.filename.

        exclusive : bool
            If True, only set if an input name is not already used
            by a parent Step. Otherwise, always set.
        """
        self._set_input_dir(obj, exclusive=exclusive)

        err_message = f"Cannot set master input file name from object {obj}"
        parent_input_filename = self.search_attr("_input_filename")
        if not exclusive or parent_input_filename is None:
            if isinstance(obj, str | Path):
                self._input_filename = str(obj)
            elif isinstance(obj, AbstractDataModel):
                try:
                    self._input_filename = obj.meta.filename
                except AttributeError:
                    self.log.debug(err_message)
            else:
                self.log.debug(err_message)

    def save_model(
        self,
        model,
        suffix=None,
        idx=None,
        output_file=None,
        force=False,
        **components,
    ):
        """
        Saves the given model using the step/pipeline's naming scheme

        Parameters
        ----------
        model : a instance of AbstractDataModel
            The model to save.

        suffix : str
            The suffix to add to the filename.

        idx : object
            Index identifier.

        output_file : str
            Use this file name instead of what the Step
            default would be.

        force : bool
            Regardless of whether ``save_results`` is `False`
            and no ``output_file`` is specified, try saving.

        components : dict
            Other components to add to the file name.

        Returns
        -------
        output_paths : [str[, ...]]
            List of output file paths the model(s) were saved in.
        """
        if output_file is None or output_file == "":
            output_file = self.output_file

        # Check if saving is even specified.
        if not force and not self.save_results and not output_file:
            return None

        if isinstance(model, AbstractModelLibrary):
            output_paths = []
            with model:
                for i, m in enumerate(model):
                    output_paths.append(
                        self.save_model(
                            m,
                            idx=i,
                            suffix=suffix,
                            force=force,
                            **components,
                        )
                    )
                    # leaving modify=True in case saving modify the file
                    model.shelve(m, i)
            return output_paths
        elif isinstance(model, Sequence):
            save_model_func = partial(
                self.save_model,
                suffix=suffix,
                force=force,
                **components,
            )
            output_path = model.save(
                path=output_file,
                save_model_func=save_model_func,
            )
        else:
            # Search for an output file name.
            if self.output_use_model or (
                output_file is None and not self.search_output_file
            ):
                output_file = model.meta.filename
                idx = None
            output_path = model.save(
                self.make_output_path(
                    basepath=output_file,
                    suffix=suffix,
                    idx=idx,
                    **components,
                )
            )
            self.log.info("Saved model in %s", output_path)

        return output_path

    @property
    def make_output_path(self):
        """Return function that creates the output path"""
        make_output_path = self.search_attr("_make_output_path")
        return partial(make_output_path, self)

    @staticmethod
    def _make_output_path(
        step,
        basepath=None,
        ext=None,
        suffix=None,
        **components,
    ):
        """Create the output path

        Parameters
        ----------
        step : Step
            The `Step` in question.

        basepath : str or None
            The basepath to use. If None, `output_file`
            is used. Only the basename component of the path
            is used.

        ext : str or None
            The extension to use. If none, `output_ext` is used.
            Can include the leading period or not.

        suffix : str or None or False
            Suffix to append to the filename.
            If None, the `Step` default will be used.
            If False, no suffix replacement will be done.

        components : dict
            dict of string replacements.

        Returns
        -------
        The fully qualified path name.

        Notes
        -----
        The values found in the `components` dict are placed in the string
        where the "{components}" replacement field is specified separated by
        underscores.
        """
        separator = "_"
        if basepath is None and step.search_output_file:
            basepath = step.search_attr("output_file")
        if basepath is None:
            basepath = step.default_output_file()

        basename, basepath_ext = splitext(split(basepath)[1])
        if ext is None:
            ext = step.output_ext
        if ext is None and len(basepath_ext):
            ext = basepath_ext
        if ext.startswith("."):
            ext = ext[1:]

        # Suffix check. An explicit check on `False` is necessary
        # because `None` is also allowed.
        suffix = _get_suffix(suffix, step=step)
        if suffix is not False:
            default_name_format = "{basename}{components}{suffix_sep}{suffix}.{ext}"
            suffix_sep = None
            if suffix is not None:
                basename, suffix_sep = step.remove_suffix(basename)
            if suffix_sep is None:
                suffix_sep = separator
        else:
            default_name_format = "{basename}{components}.{ext}"
            suffix = None
            suffix_sep = None

        # Setup formatting
        formatter = FormatTemplate(
            separator=separator,
            remove_unused=True,
        )

        if len(components):
            component_str = formatter("", **components)
        else:
            component_str = ""

        basename = formatter(
            default_name_format,
            basename=basename,
            suffix=suffix,
            suffix_sep=suffix_sep,
            ext=ext,
            components=component_str,
        )

        output_dir = step.search_attr("output_dir", default="")
        output_dir = expandvars(expanduser(output_dir))
        return join(output_dir, basename)

    @classmethod
    def _datamodels_open(cls, init, **kwargs):
        """
        Wrapper around observatory-specific datamodels.open function.
        """
        raise NotImplementedError(f"{cls.__name__} does not implement _datamodels_open")

    def open_model(self, init, **kwargs):
        """Open a datamodel

        Primarily a wrapper around ``DataModel.open`` to
        handle `Step` peculiarities

        Parameters
        ----------
        init : object
            The object to open

        Returns
        -------
        datamodel : instance of AbstractDataModel
            Object opened as a datamodel
        """
        # Use the parent method if available, since this step
        # might be a hook that doesn't implement _datamodels_open.
        if self.parent is None:
            datamodels_open = self._datamodels_open
        else:
            datamodels_open = self.parent._datamodels_open

        return datamodels_open(self.make_input_path(init), **kwargs)

    def make_input_path(self, file_path):
        """Create an input path for a given file path

        If ``file_path`` has no directory path, use ``self.input_dir``
        as the directory path.

        Parameters
        ----------
        file_path : str or obj
            The supplied file path to check and modify.
            If anything other than `str`, the object
            is simply passed back.

        Returns
        -------
        full_path : str or obj
            File path using ``input_dir`` if the input
            had no directory path.
        """
        full_path = file_path
        if isinstance(file_path, str):
            original_path, file_name = split(file_path)
            if not len(original_path):
                full_path = join(self.input_dir, file_name)

        return full_path

    def _set_input_dir(self, input_, exclusive=True):
        """Set the input directory

        If sufficient information is at hand, set a value
        for the attribute `input_dir`.

        Parameters
        ----------
        input_ : str
            Input to determine path from.

        exclusive : bool
            If True, only set if an input directory is not already
            defined by a parent Step. Otherwise, always set.

        """
        if not exclusive or self.search_attr("_input_dir") is None:
            with suppress(Exception):
                if isfile(input_):
                    self.input_dir = split(input_)[0]

    def get_pars(self, full_spec=True):
        """Retrieve the configuration parameters of a step

        Parameters
        ----------
        full_spec : bool
            Return all parameters, including parent-specified parameters.
            If `False`, return only parameters specific to the step.

        Returns
        -------
        dict
            Keys are the parameters and values are the values.
        """
        from . import cmdline

        if full_spec:
            spec_file_func = config_parser.get_merged_spec_file
        else:
            spec_file_func = config_parser.load_spec_file
        spec = spec_file_func(self)
        if spec is None:
            return {}
        instance_pars = {}
        for key in spec:
            if hasattr(self, key):
                value = getattr(self, key)
                instance_pars[key] = value
        pars = config_parser.config_from_dict(instance_pars, spec, allow_missing=True)

        # Convert the config to a pure dict.
        pars_dict = {}
        for key, value in pars.items():
            if isinstance(value, cmdline.FromCommandLine):
                pars_dict[key] = str(value)
            else:
                pars_dict[key] = value
        return pars_dict

    def export_config(self, filename, include_metadata=False):
        """
        Export this step's parameters to an ASDF config file.

        Parameters
        ----------
        filename : str or pathlib.Path
            Path to config file.

        include_metadata : bool, optional
            Set to True to include metadata that is required
            for submission to CRDS.
        """
        with config.export_config(self).to_asdf(
            include_metadata=include_metadata
        ) as af:
            af.write_to(filename)

    def update_pars(self, parameters):
        """Update step parameters

        Only existing parameters are updated. Otherwise, new keys
        found in ``parameters`` are ignored.

        Parameters
        ----------
        parameters : dict
            Parameters to update.

        Notes
        -----
        ``parameters`` is presumed to have been produced by the
        `Step.get_pars` method. As such, the "steps" key is treated
        special in that it is a dict whose keys are the steps assigned
        directly as parameters to the current step. This is standard
        practice for `Pipeline`-based steps.
        """
        existing = self.get_pars().keys()
        for parameter, value in parameters.items():
            if parameter in existing:
                if parameter != "steps":
                    setattr(self, parameter, value)
                else:
                    for step_name, step_parameters in value.items():
                        getattr(self, step_name).update_pars(step_parameters)
            else:
                self.log.debug(
                    "Parameter %s is not valid for step %s. Ignoring.", parameter, self
                )

    @classmethod
    def build_config(cls, input, **kwargs):  # noqa: A002
        """Build the ConfigObj to initialize a Step

        A Step config is built in the following order:

        - CRDS parameter reference file
        - Local parameter reference file
        - Step keyword arguments

        Parameters
        ----------
        input : str or None
            Input file

        kwargs : dict
            Keyword arguments that specify Step parameters.

        Returns
        -------
        config, config_file : ConfigObj, str
            The configuration and the config filename.
        """
        logger_name = cls.__name__
        log_cls = log.getLogger(logger_name)
        if input:
            config = cls.get_config_from_reference(input)
        else:
            log_cls.info("No filename given, cannot retrieve config from CRDS")
            config = config_parser.ConfigObj()

        if "config_file" in kwargs:
            config_file = kwargs["config_file"]
            del kwargs["config_file"]
            config_from_file = config_parser.load_config_file(str(config_file))
            config_parser.merge_config(config, config_from_file)
            config_dir = os.path.dirname(config_file)
        else:
            config_file = None
            config_dir = ""

        config_kwargs = config_parser.ConfigObj()

        # load and merge configuration files for each step they are provided:
        steps = {}
        if "steps" in kwargs:
            for step, pars in kwargs["steps"].items():
                if "config_file" in pars:
                    step_config_file = os.path.join(config_dir, pars["config_file"])
                    cfgd = config_parser.load_config_file(step_config_file)
                    if "name" in cfgd:
                        if cfgd["name"] != step:
                            raise ValueError(
                                "Step name from configuration file "
                                f"'{step_config_file}' does not match step "
                                "name in the 'steps' argument."
                            )
                        del cfgd["name"]
                    cfgd.pop("class", None)
                    cfgd.update(pars)
                    steps[step] = cfgd
                else:
                    steps[step] = pars

            kwargs = {k: v for k, v in kwargs.items() if k != "steps"}
            if steps:
                kwargs["steps"] = steps

        config_parser.merge_config(config_kwargs, kwargs)
        config_parser.merge_config(config, config_kwargs)

        return config, config_file


# #########
# Utilities
# #########


def _get_suffix(suffix, step=None, default_suffix=None):
    """Retrieve either specified or pipeline-supplied suffix

    Parameters
    ----------
    suffix : str or None
        Suffix to use if specified.

    step : Step or None
        The step to retrieve the suffux.

    default_suffix : str
        If the pipeline does not supply a suffix,
        use this.

    Returns
    -------
    suffix : str or None
        Suffix to use
    """
    if suffix is None and step is not None:
        suffix = step.search_attr("suffix")
    if suffix is None:
        suffix = default_suffix
    if suffix is None and step is not None:
        suffix = step.name.lower()
    return suffix


def get_disable_crds_steppars(default=None):
    """Return either the explicit default flag or retrieve from the environment

    If a default is not specified, retrieve the value from the environmental variable
    `STPIPE_DISABLE_CRDS_STEPPARS`.

    Parameters
    ----------
    default: str, bool, or None
        Flag to use. If None, the environmental is used.

    Returns
    -------
    flag: bool
        True to disable CRDS STEPPARS retrieval.
    """
    truths = ("true", "True", "t", "yes", "y")
    if default:
        if isinstance(default, bool):
            return default

        if isinstance(default, str):
            return default in truths

        raise ValueError(f"default must be string or boolean: {default}")

    flag = os.environ.get("STPIPE_DISABLE_CRDS_STEPPARS", "")
    return flag in truths


@contextmanager
def preserve_step_pars(step):
    """Context manager to preserve step parameters

    Ensure step parameters are not modified during a block
    of operations. Allows local reuse of a Step instance without
    having to worry about side-effects on that Step. If used with
    a `Pipeline`, all substep parameters are also restored.

    Yields
    ------
    saved_pars: dict
        The saved parameters.
    """
    saved_pars = step.get_pars()
    try:
        yield saved_pars
    finally:
        step.update_pars(saved_pars)


"""
This module is designed to make romancal tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


romancal copyright notice:

Copyright (C) 2010 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/romancal/blob/main/romancal/associations/load_asn.py
"""

from inspect import isclass

from spike.romancal.associations_association import Association
from spike.romancal.associations_registry import AssociationRegistry


def load_asn(
    serialized,
    format=None,
    first=True,
    validate=True,
    registry=AssociationRegistry,
    **kwargs,
):
    """Load an Association from a file or object

    Parameters
    ----------
    serialized : object
        The serialized form of the association.

    format : str or None
        The format to force. If None, try all available.

    validate : bool
        Validate against the class' defined schema, if any.

    first : bool
        A serialization potentially matches many rules.
        Only return the first succesful load.

    registry : AssociationRegistry or None
        The `AssociationRegistry` to use.
        If None, no registry is used.
        Can be passed just a registry class instead of instance.

    kwargs : dict
        Other arguments to pass to the `load` methods defined
        in the `Association.IORegistry`

    Returns
    -------
    The Association object

    Raises
    ------
    AssociationNotValidError
        Cannot create or validate the association.

    Notes
    -----
    The `serialized` object can be in any format
    supported by the registered I/O routines. For example, for
    `json` and `yaml` formats, the input can be either a string or
    a file object containing the string.

    If no registry is specified, the default `Association.load`
    method is used.
    """
    if registry is None:
        return Association.load(serialized, format=format, validate=validate)

    if isclass(registry):
        registry = registry()
    return registry.load(
        serialized, format=format, first=first, validate=validate, **kwargs
    )


"""
This module is designed to make jwst tweakreg and resample functions accessible without
installing the original package due to their complex dependencies. As such, it is only subtly modified from
the original to accommodate the less stringent install requirements.


jwst copyright notice:

Copyright (C) 2020 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.

Original: https://github.com/spacetelescope/jwst/blob/main/jwst/stpipe/utilities.py

"""

# Copyright (C) 2010 Association of Universities for Research in Astronomy(AURA)
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
#     1. Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#
#     2. Redistributions in binary form must reproduce the above
#       copyright notice, this list of conditions and the following
#       disclaimer in the documentation and/or other materials provided
#       with the distribution.
#
#     3. The name of AURA and its representatives may not be used to
#       endorse or promote products derived from this software without
#       specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
# WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
# MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
# OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
# TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
# USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
# DAMAGE.
"""
Utilities
"""
import importlib.util
from importlib import import_module
import inspect
import logging
import os
import re
from collections.abc import Sequence
# from jwst import datamodels
from spike.jwstcal.datamodels_library import ModelLibrary

# Configure logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

# Step classes that are not user-api steps
NON_STEPS = [
    'EngDBLogStep',
    'FunctionWrapper',
    'JwstPipeline',
    'JwstStep',
    'Pipeline',
    'Step',
    'SystemCall',
]

NOT_SET = "NOT SET"
COMPLETE = "COMPLETE"
SKIPPED = "SKIPPED"

def all_steps():
    """List all classes subclassed from Step

    Returns
    -------
    steps : dict
        Key is the classname, value is the class
    """
    from spike.jwstcal.stpipe_core import JwstStep as Step

    jwst = import_module('jwst')
    jwst_fpath = os.path.split(jwst.__file__)[0]

    steps = {}
    for module in load_local_pkg(jwst_fpath):
        more_steps = {
            klass_name: klass
            for klass_name, klass in inspect.getmembers(
                module,
                lambda o: inspect.isclass(o) and issubclass(o, Step)
            )
            if klass_name not in NON_STEPS
        }
        steps.update(more_steps)

    return steps


def load_local_pkg(fpath):
    """Generator producing all modules under fpath

    Parameters
    ----------
    fpath: string
        File path to the package to load.

    Returns
    -------
    generator
        `module` for each module found in the package.
    """
    package_fpath, package = os.path.split(fpath)
    package_fpath_len = len(package_fpath) + 1
    try:
        for module_fpath in folder_traverse(
            fpath, basename_regex=r'[^_].+\.py$', path_exclude_regex='(tests)|(regtest)'
        ):
            folder_path, fname = os.path.split(module_fpath[package_fpath_len:])
            module_path = folder_path.split('/')
            module_path.append(os.path.splitext(fname)[0])
            module_path = '.'.join(module_path)
            try:
                spec = importlib.util.spec_from_file_location(module_path, module_fpath)
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
            except Exception as err:
                logger.debug(f'Cannot load module "{module_path}": {str(err)}')
            else:
                yield module
    except Exception as err:
        logger.debug(f'Cannot complete package loading: Exception occurred: "{str(err)}"')


def folder_traverse(folder_path, basename_regex='.+', path_exclude_regex='^$'):
    """Generator of full file paths for all files
    in a folder.

    Parameters
    ----------
    folder_path: str
        The folder to traverse

    basename_regex: str
        Regular expression that must match
        the `basename` part of the file path.

    path_exclude_regex: str
        Regular expression to exclude a path.

    Returns
    -------
    generator
        A generator, return the next file.
    """
    basename_regex = re.compile(basename_regex)
    path_exclude_regex = re.compile(path_exclude_regex)
    for root, dirs, files in os.walk(folder_path):
        if path_exclude_regex.search(root):
            continue
        for file in files:
            if basename_regex.match(file):
                yield os.path.join(root, file)


def record_step_status(datamodel, cal_step, success=True):
    """Record whether or not a step completed in meta.cal_step

    Parameters
    ----------
    datamodel : `~jwst.datamodels.JwstDataModel`, `~jwst.datamodels.ModelContainer`, `~jwst.datamodels.ModelLibrary`, str, or Path instance
        This is the datamodel or container of datamodels to modify in place

    cal_step : str
        The attribute in meta.cal_step for recording the status of the step

    success : bool
        If True, then 'COMPLETE' is recorded.  If False, then 'SKIPPED'
    """
    if success:
        status = COMPLETE
    else:
        status = SKIPPED

    if isinstance(datamodel, Sequence):
        for model in datamodel:
            model.meta.cal_step._instance[cal_step] = status
    elif isinstance(datamodel, ModelLibrary): #modified path
        with datamodel:
            for model in datamodel:
                model.meta.cal_step._instance[cal_step] = status
                datamodel.shelve(model)
    else:
        datamodel.meta.cal_step._instance[cal_step] = status

    # TODO: standardize cal_step naming to point to the official step name


def query_step_status(datamodel, cal_step):
    """Query the status of a step in meta.cal_step
    
    Parameters
    ----------
    datamodel : `~jwst.datamodels.JwstDataModel` or `~jwst.datamodels.ModelContainer` instance
        The datamodel or container of datamodels to check
        
    cal_step : str
        The attribute in meta.cal_step to check
    
    Returns
    -------
    status : str
        The status of the step in meta.cal_step, typically 'COMPLETE' or 'SKIPPED'

    Notes
    -----
    In principle, a step could set the COMPLETE status for only some subset 
    of models, so checking the zeroth model instance may not always be correct.
    However, this is not currently done in the pipeline. This function should be
    updated to accommodate that use-case as needed.
    """
    if isinstance(datamodel, Sequence):
        return getattr(datamodel[0].meta.cal_step, cal_step, NOT_SET)
    else:
        return getattr(datamodel.meta.cal_step, cal_step, NOT_SET)